# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the admin service implementation for v1."""

import warnings

import grpc
import semver

from ansys.geometry.core.errors import protect_grpc

from ..base.admin import GRPCAdminService
from .conversions import (
    build_grpc_id,
    from_grpc_backend_type_to_backend_type,
    serialize_tracked_changes,
)


class GRPCAdminServiceV1(GRPCAdminService):
    """Admin service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    admin service. It is specifically designed for the v1 version of the
    Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.discovery.v1.commands.application_pb2_grpc import ApplicationStub
        from ansys.api.discovery.v1.commands.communication_pb2_grpc import CommunicationStub

        self.stub = ApplicationStub(channel)
        self.communication_stub = CommunicationStub(channel)

    @protect_grpc
    def get_backend(self, **kwargs) -> dict:  # noqa: D102
        # TODO: Remove this context and filter once the protobuf UserWarning is downgraded to INFO
        # https://github.com/grpc/grpc/issues/37609
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "Protobuf gencode version", UserWarning, "google.protobuf.runtime_version"
            )
            from ansys.api.discovery.v1.commands.application_pb2 import GetBackendRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetBackendRequest()

        # Call the gRPC service
        response = self.stub.GetBackend(request=request)

        ver = response.version
        backend_version = semver.Version(ver.major_release, ver.minor_release, ver.service_pack)
        api_server_build_info = f"{ver.build_number}" if ver.build_number != 0 else "N/A"
        product_build_info = (
            response.backend_version_info.strip() if response.backend_version_info else "N/A"
        )

        # Convert the response to a dictionary
        return {
            "backend": from_grpc_backend_type_to_backend_type(response.type),
            "version": backend_version,
            "api_server_build_info": api_server_build_info,
            "product_build_info": product_build_info,
            "additional_info": {k: v for k, v in response.additional_build_info.items()},
        }

    @protect_grpc
    def get_logs(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commands.communication_pb2 import LogsRequest
        from ansys.api.discovery.v1.commonenums_pb2 import LogsPeriodType, LogsTarget

        # Create the request - assumes all inputs are valid and of the proper type
        request = LogsRequest(
            target=LogsTarget.LOGSTARGET_CLIENT,
            period_type=(
                LogsPeriodType.LOGSPERIODTIME_CURRENT
                if not kwargs["all_logs"]
                else LogsPeriodType.LOGSPERIODTIME_ALL
            ),
            null_path=None,
            null_period=None,
        )

        # Call the gRPC service
        logs_generator = self.communication_stub.GetLogs(request)
        logs: dict[str, str] = {}

        # Convert the response to a dictionary
        for chunk in logs_generator:
            if chunk.log_name not in logs:
                logs[chunk.log_name] = ""
            logs[chunk.log_name] += chunk.log_chunk.decode()

        return {"logs": logs}

    @protect_grpc
    def get_service_status(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commands.communication_pb2 import HealthRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = HealthRequest()

        # Call the gRPC service
        response = self.communication_stub.Health(request=request)

        # Convert the response to a dictionary
        return {"healthy": True if response.message == "I am healthy!" else False}

    @protect_grpc
    def get_tracker(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commands.application_pb2 import GetTrackerRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetTrackerRequest(design_id=build_grpc_id(kwargs["design_id"]))

        # Call the gRPC service
        response = self.stub.GetTracker(request=request)

        # Convert the response to a dictionary
        return {"design_id": response.tracker.design_id.id}

    @protect_grpc
    def get_tracker_changes(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commands.application_pb2 import GetTrackerChangesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetTrackerChangesRequest(design_id=build_grpc_id(kwargs["design_id"]))

        # Call the gRPC service
        response = self.stub.GetTrackerChanges(request=request)

        # Convert the response to a dictionary
        return serialize_tracked_changes(response.tracked_changes)

    @protect_grpc
    def set_automatic_tracking_state(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commands.application_pb2 import SetAutomaticTrackingStateRequest
        from ansys.api.discovery.v1.commonenums_pb2 import AutomaticTrackingState

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetAutomaticTrackingStateRequest(
            state=(
                AutomaticTrackingState.AUTOMATICTRACKINGSTATE_ON
                if kwargs["enabled"]
                else AutomaticTrackingState.AUTOMATICTRACKINGSTATE_OFF
            ),
            design_id=build_grpc_id(kwargs["design_id"]),
        )

        # Call the gRPC service
        response = self.stub.SetAutomaticTrackingState(request=request)

        # Convert the response to a dictionary
        return {"success": response.set_response.success}



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the designs service implementation for v1."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.designs import GRPCDesignsService
from .conversions import (
    build_grpc_id,
    from_grpc_curve_to_curve,
    from_grpc_frame_to_frame,
    from_grpc_material_to_material,
    from_grpc_matrix_to_matrix,
    from_grpc_plane_to_plane,
    from_grpc_point_to_point3d,
)


class GRPCDesignsServiceV1(GRPCDesignsService):
    """Designs service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    designs service. It is specifically designed for the v1 version of the
    Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.discovery.v1.commands.file_pb2_grpc import FileStub
        from ansys.api.discovery.v1.design.designdoc_pb2_grpc import DesignDocStub

        self.file_stub = FileStub(channel)
        self.designdoc_stub = DesignDocStub(channel)

    @protect_grpc
    def open(self, **kwargs) -> dict:  # noqa: D102
        from pathlib import Path
        from typing import TYPE_CHECKING, Generator

        from ansys.api.discovery.v1.commands.file_pb2 import OpenMode, OpenRequest

        import ansys.geometry.core.connection.defaults as pygeom_defaults

        from .conversions import from_import_options_definitions_to_grpc_import_options_definition

        if TYPE_CHECKING:  # pragma: no cover
            from ansys.geometry.core.misc.options import ImportOptions, ImportOptionsDefinitions

        def request_generator(
            file_path: Path,
            file_name: str,
            import_options: "ImportOptions",
            import_options_definitions: "ImportOptionsDefinitions",
            open_mode: OpenMode,
        ) -> Generator[OpenRequest, None, None]:
            """Generate requests for streaming file upload."""
            msg_buffer = 5 * 1024  # 5KB - for additional message data
            if pygeom_defaults.MAX_MESSAGE_LENGTH - msg_buffer < 0:  # pragma: no cover
                raise ValueError("MAX_MESSAGE_LENGTH is too small for file upload.")

            chunk_size = pygeom_defaults.MAX_MESSAGE_LENGTH - msg_buffer
            with Path.open(file_path, "rb") as file:
                while chunk := file.read(chunk_size):
                    test_req = OpenRequest(
                        data=chunk,
                        file_name=file_name,
                        open_mode=open_mode,
                        import_options=import_options.to_dict(),
                        import_options_definitions=from_import_options_definitions_to_grpc_import_options_definition(
                            import_options_definitions
                        ),
                    )
                    yield test_req

        # Get the open mode
        open_mode = kwargs["open_mode"]
        if open_mode == "new":
            open_mode = OpenMode.OPENMODE_NEW
        elif open_mode == "insert":
            open_mode = OpenMode.OPENMODE_INSERT

        # Call the gRPC service
        response = self.file_stub.Open(
            request_generator(
                file_path=kwargs["filepath"],
                file_name=kwargs["original_file_name"],
                import_options=kwargs["import_options"],
                import_options_definitions=kwargs["import_options_definitions"],
                open_mode=open_mode,
            )
        )

        # Return the response - formatted as a dictionary
        return {"file_path": response.design.path}

    @protect_grpc
    def new(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commands.file_pb2 import NewRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = NewRequest(name=kwargs["name"])

        # Call the gRPC service
        response = self.file_stub.New(request)

        # Return the response - formatted as a dictionary
        return {
            "design_id": response.design.id.id,
            "main_part_id": response.design.main_part_id.id,
        }

    @protect_grpc
    def get_assembly(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.designdoc_pb2 import GetAssemblyRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetAssemblyRequest(id=build_grpc_id(kwargs["active_design"].get("design_id")))

        # Call the gRPC service
        response = self.designdoc_stub.GetAssembly(request)

        # Return the response - formatted as a dictionary
        serialized_response = self._serialize_assembly_response(response)
        return serialized_response

    @protect_grpc
    def close(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commands.file_pb2 import CloseRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = CloseRequest(design_id=build_grpc_id(kwargs["design_id"]))

        # Call the gRPC service
        _ = self.file_stub.Close(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def put_active(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError(
            f"Method '{self.__class__.__name__}.put_active' is not "
            "implemented in this protofile version."
        )

    @protect_grpc
    def save_as(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commands.file_pb2 import SaveRequest

        from .conversions import (
            _check_write_body_facets_input,
            from_design_file_format_to_grpc_file_export_format,
        )

        _check_write_body_facets_input(kwargs["backend_version"], kwargs["write_body_facets"])

        # Create the request - assumes all inputs are valid and of the proper type
        request = SaveRequest(
            format=from_design_file_format_to_grpc_file_export_format(kwargs["format"]),
            write_body_facets=kwargs["write_body_facets"],
        )

        # Call the gRPC service
        response_stream = self.file_stub.Save(request)

        # Return the response - formatted as a dictionary
        data = bytes()
        for response in response_stream:
            data += response.data

        return {
            "data": data,
        }

    @protect_grpc
    def download_export(self, **kwargs) -> dict:  # noqa: D102
        return self.save_as(**kwargs)

    @protect_grpc
    def stream_download_export(self, **kwargs) -> dict:  # noqa: D102
        return self.save_as(**kwargs)

    @protect_grpc
    def insert(self, **kwargs) -> dict:  # noqa: D102
        # Route to open method
        return self.open(**kwargs, open_mode="insert")

    @protect_grpc
    def get_active(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import EntityRequest

        # Call the gRPC service
        response = self.designdoc_stub.Get(request=EntityRequest(id=build_grpc_id("")))

        # Return the response - formatted as a dictionary
        if response.design:
            return {
                "design_id": response.design.id.id,
                "main_part_id": response.design.main_part_id.id,
                "name": response.design.name,
            }

    @protect_grpc
    def upload_file(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError(
            f"Method '{self.__class__.__name__}.upload_file' is not "
            "implemented in this protofile version."
        )

    @protect_grpc
    def upload_file_stream(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError(
            f"Method '{self.__class__.__name__}.upload_file_stream' is not "
            "implemented in this protofile version."
        )

    @protect_grpc
    def stream_design_tessellation(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.designdoc_pb2 import DesignTessellationRequest

        from .conversions import (
            from_grpc_edge_tess_to_raw_data,
            from_grpc_tess_to_raw_data,
            from_tess_options_to_grpc_tess_options,
        )

        # If there are options, convert to gRPC options
        options = (
            from_tess_options_to_grpc_tess_options(kwargs["options"])
            if kwargs["options"] is not None
            else None
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = DesignTessellationRequest(
            options=options,
            include_faces=kwargs["include_faces"],
            include_edges=kwargs["include_edges"],
        )

        # Call the gRPC service
        response_stream = self.designdoc_stub.StreamDesignTessellation(request)

        # Return the response - formatted as a dictionary
        tess_map = {}
        for elem in response_stream:
            for body_id, body_tess in elem.body_tessellation.items():
                tess = {}
                for face_id, face_tess in body_tess.face_tessellation.items():
                    tess[face_id] = from_grpc_tess_to_raw_data(face_tess)
                for edge_id, edge_tess in body_tess.edge_tessellation.items():
                    tess[edge_id] = from_grpc_edge_tess_to_raw_data(edge_tess)
                tess_map[body_id] = tess

        return {
            "tessellation": tess_map,
        }

    @protect_grpc
    def download_file(self, **kwargs) -> dict:  # noqa: D102
        # This method is only accessed by __export_and_download_legacy,
        # which is only used for older versions of the server. Since this class
        # is only used for v1 of the protofiles, this method will not be implemented
        # here. If this method is called, it means there is a mismatch between the server
        # version and the protofiles version being used, and an error should be raised to
        # alert the user of this mismatch.
        raise NotImplementedError(  # pragma: no cover
            f"Method '{self.__class__.__name__}.download_file' is not "
            "implemented in this protofile version."
        )

    def _serialize_assembly_response(self, response):
        def serialize_body(body):
            return {
                "id": body.id.id,
                "name": body.name,
                "master_id": body.master_id.id,
                "parent_id": body.parent_id.id,
                "is_surface": body.is_surface,
            }

        def serialize_component(component):
            return {
                "id": component.id.id,
                "parent_id": component.parent_id.id,
                "master_id": component.master_id.id,
                "name": component.name,
                "placement": component.placement,
                "part_master": serialize_part(component.part_master),
            }

        def serialize_transformed_part(transformed_part):
            return {
                "id": transformed_part.id.id,
                "name": transformed_part.name,
                "placement": from_grpc_matrix_to_matrix(transformed_part.placement),
                "part_master": serialize_part(transformed_part.part_master),
            }

        def serialize_part(part):
            return {
                "id": part.id.id,
                "name": part.name,
            }

        def serialize_material_properties(material_property):
            return {
                "id": material_property.id.id,
                "display_name": material_property.display_name,
                "value": material_property.value,
                "units": material_property.units,
            }

        def serialize_material(material):
            material_properties = getattr(material, "material_properties", [])
            return {
                "name": material.name,
                "material_properties": [
                    serialize_material_properties(property) for property in material_properties
                ],
            }

        def serialize_named_selection(named_selection):
            return {"id": named_selection.id.id, "name": named_selection.name}

        def serialize_coordinate_systems(coordinate_systems):
            serialized_cs = []
            for cs in coordinate_systems.coordinate_systems:
                serialized_cs.append(
                    {
                        "id": cs.id.id,
                        "name": cs.name,
                        "frame": from_grpc_frame_to_frame(cs.frame),
                    }
                )

            return serialized_cs

        def serialize_component_coordinate_systems(component_coordinate_system):
            serialized_component_coordinate_systems = []
            for (
                component_coordinate_system_id,
                coordinate_systems,
            ) in component_coordinate_system.items():
                serialized_component_coordinate_systems.append(
                    {
                        "component_id": component_coordinate_system_id,
                        "coordinate_systems": serialize_coordinate_systems(coordinate_systems),
                    }
                )

            return serialized_component_coordinate_systems

        def serialize_component_shared_topologies(component_share_topology):
            serialized_share_topology = []
            for component_shared_topology_id, shared_topology in component_share_topology.items():
                serialized_share_topology.append(
                    {
                        "component_id": component_shared_topology_id,
                        "shared_topology_type": shared_topology,
                    }
                )
            return serialized_share_topology

        def serialize_beam_curve(curve):
            return {
                "curve": from_grpc_curve_to_curve(curve.curve),
                "start": from_grpc_point_to_point3d(curve.start),
                "end": from_grpc_point_to_point3d(curve.end),
                "interval_start": curve.interval_start,
                "interval_end": curve.interval_end,
                "length": curve.length,
            }

        def serialize_beam_curve_list(curve_list):
            return {"curves": [serialize_beam_curve(curve) for curve in curve_list.curves]}

        def serialize_beam_cross_section(cross_section):
            return {
                "section_anchor": cross_section.section_anchor,
                "section_angle": cross_section.section_angle.value_in_geometry_units,
                "section_frame": from_grpc_frame_to_frame(cross_section.section_frame),
                "section_profile": [
                    serialize_beam_curve_list(curve_list)
                    for curve_list in cross_section.section_profile
                ],
            }

        def serialize_beam_properties(properties):
            return {
                "area": properties.area.value_in_geometry_units,
                "centroid_x": properties.centroid_x.value_in_geometry_units,
                "centroid_y": properties.centroid_y.value_in_geometry_units,
                "warping_constant": properties.warping_constant.value_in_geometry_units,
                "ixx": properties.ixx.value_in_geometry_units,
                "ixy": properties.ixy.value_in_geometry_units,
                "iyy": properties.iyy.value_in_geometry_units,
                "shear_center_x": properties.shear_center_x.value_in_geometry_units,
                "shear_center_y": properties.shear_center_y.value_in_geometry_units,
                "torsional_constant": properties.torsional_constant.value_in_geometry_units,
            }

        def serialize_beam(beam):
            return {
                "id": beam.id.id,
                "parent_id": beam.parent.id,
                "start": from_grpc_point_to_point3d(beam.shape.start),
                "end": from_grpc_point_to_point3d(beam.shape.end),
                "name": beam.name,
                "is_deleted": beam.is_deleted,
                "is_reversed": beam.is_reversed,
                "is_rigid": beam.is_rigid,
                "material": from_grpc_material_to_material(beam.material),
                "type": beam.type,
                "properties": serialize_beam_properties(beam.properties),
                "cross_section": serialize_beam_cross_section(beam.cross_section),
            }

        def serialize_design_point(design_point):
            return {
                "id": design_point.id.id,
                "name": design_point.owner_name,
                "point": from_grpc_point_to_point3d(design_point.points[0]),
                "parent_id": design_point.parent_id.id,
            }

        def serialize_datum_plane(datum_plane):
            return {
                "id": datum_plane.id.id,
                "name": datum_plane.name,
                "plane": from_grpc_plane_to_plane(datum_plane.plane),
                "parent_id": datum_plane.parent_id.id,
            }

        parts = getattr(response, "parts", [])
        transformed_parts = getattr(response, "transformed_parts", [])
        bodies = getattr(response, "bodies", [])
        components = getattr(response, "components", [])
        materials = getattr(response, "materials", [])
        named_selections = getattr(response, "named_selections", [])
        component_coordinate_systems = getattr(response, "component_coord_systems", [])
        component_shared_topologies = getattr(response, "component_shared_topologies", [])
        beams = getattr(response, "beams", [])
        design_points = getattr(response, "design_points", [])
        datum_planes = getattr(response, "datum_planes", [])
        return {
            "parts": [serialize_part(part) for part in parts] if len(parts) > 0 else [],
            "transformed_parts": [serialize_transformed_part(tp) for tp in transformed_parts],
            "bodies": [serialize_body(body) for body in bodies] if len(bodies) > 0 else [],
            "components": [serialize_component(component) for component in components],
            "materials": [serialize_material(material) for material in materials],
            "named_selections": [serialize_named_selection(ns) for ns in named_selections],
            "component_coordinate_systems": serialize_component_coordinate_systems(
                component_coordinate_systems
            ),
            "component_shared_topologies": serialize_component_shared_topologies(
                component_shared_topologies
            ),
            "beams": [serialize_beam(beam) for beam in beams],
            "design_points": [serialize_design_point(dp) for dp in design_points],
            "datum_planes": [serialize_datum_plane(dp) for dp in datum_planes],
        }



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Auxiliary functions for the PyAnsys Geometry library."""

from pathlib import Path
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.beam import Beam
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.component import Component
    from ansys.geometry.core.designer.design import Design
    from ansys.geometry.core.designer.designpoint import DesignPoint
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face
    from ansys.geometry.core.designer.vertex import Vertex

try:
    from ansys.tools.visualization_interface.utils.color import Color

    DEFAULT_COLOR = Color.DEFAULT.value
except ModuleNotFoundError:
    DEFAULT_COLOR = "#D6F7D1"


def get_design_from_component(component: "Component") -> "Design":
    """Get the ``Design`` of the given ``Component`` object.

    Parameters
    ----------
    component : Component
        The component object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided component object.
    """
    # Get the design of the component
    while component.parent_component is not None:
        component = component.parent_component

    # Return the design
    return component


def get_design_from_body(body: "Body") -> "Design":
    """Get the ``Design`` of the given ``Body`` object.

    Parameters
    ----------
    body : Body
        The body object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided body object.
    """
    # Get the parent component of the body
    component = body.parent_component

    # Get the design of the component
    return get_design_from_component(component)


def get_design_from_face(face: "Face") -> "Design":
    """Get the ``Design`` of the given ``Face`` object.

    Parameters
    ----------
    face : Face
        The face object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided face object.
    """
    # Get the parent body of the face
    body = face.body

    # Get the design of the body
    return get_design_from_body(body)


def get_design_from_edge(edge: "Edge") -> "Design":
    """Get the ``Design`` of the given ``Edge`` object.

    Parameters
    ----------
    edge : Edge
        The edge object for which to find the ``Design``.

    Returns
    -------
    Design
        The ``Design`` of the provided edge object.
    """
    # Get the one of the bodies of the edge
    body = edge.faces[0].body

    # Get the design of the body
    return get_design_from_body(body)


def __traverse_component_elem(elem: str, comp: Union["Design", "Component"]) -> list:
    """Traverses all elements of a component or design, given its name."""
    elems = []
    comp_elems = getattr(comp, elem)
    elems.extend(comp_elems)
    for component in comp.components:
        elems.extend(__traverse_component_elem(elem, component))

    return elems


def __traverse_all_bodies(comp: Union["Design", "Component"]) -> list["Body"]:
    """Traverse all bodies in a design/component and all its subcomponents."""
    return __traverse_component_elem("bodies", comp)


def __traverse_all_beams(comp: Union["Design", "Component"]) -> list["Body"]:
    """Traverse all beams in a design/component and all its subcomponents."""
    return __traverse_component_elem("beams", comp)


def __traverse_all_design_points(comp: Union["Design", "Component"]) -> list["DesignPoint"]:
    """Traverse all design points in a design/component and all its subcomponents."""
    return __traverse_component_elem("design_points", comp)


def get_all_bodies_from_design(design: "Design") -> list["Body"]:
    """Find all the ``Body`` objects inside a ``Design``.

    Parameters
    ----------
    design : Design
        Parent design for the bodies.

    Returns
    -------
    list[Body]
        List of Body objects.

    Notes
    -----
    This method takes a design and gets the corresponding ``Body`` objects.
    """
    return __traverse_all_bodies(design)


def get_bodies_from_ids(design: "Design", body_ids: list[str]) -> list["Body"]:
    """Find the ``Body`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the faces.
    body_ids : list[str]
        List of body ids.

    Returns
    -------
    list[Body]
        List of Body objects.

    Notes
    -----
    This method takes a design and body ids, and gets their corresponding ``Body`` object.
    """
    return [body for body in __traverse_all_bodies(design) if body.id in body_ids]


def get_components_from_ids(design: "Design", component_ids: list[str]) -> list["Component"]:
    """Find the ``Component`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the components.
    component_ids : list[str]
        List of component ids.

    Returns
    -------
    list[Component]
        List of Component objects.

    Notes
    -----
    This method takes a design and component ids, and gets their corresponding ``Component`` object.
    """
    return [
        comp for comp in __traverse_component_elem("components", design) if comp.id in component_ids
    ]  # noqa: E501


def get_faces_from_ids(design: "Design", face_ids: list[str]) -> list["Face"]:
    """Find the ``Face`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the faces.
    face_ids : list[str]
        List of face ids.

    Returns
    -------
    list[Face]
        List of Face objects.

    Notes
    -----
    This method takes a design and face ids, and gets their corresponding ``Face`` object.
    """
    return [
        face for body in __traverse_all_bodies(design) for face in body.faces if face.id in face_ids
    ]  # noqa: E501


def get_edges_from_ids(design: "Design", edge_ids: list[str]) -> list["Edge"]:
    """Find the ``Edge`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the edges.
    edge_ids : list[str]
        List of edge ids.

    Returns
    -------
    list[Edge]
        List of Edge objects.

    Notes
    -----
    This method takes a design and edge ids, and gets their corresponding ``Edge`` objects.
    """
    return [
        edge for body in __traverse_all_bodies(design) for edge in body.edges if edge.id in edge_ids
    ]  # noqa: E501


def get_vertices_from_ids(design: "Design", vertex_ids: list[str]) -> list["Vertex"]:
    """Find the ``Vertex`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the vertices.
    vertex_ids : list[str]
        List of vertex ids.

    Returns
    -------
    list[Vertex]
        List of Vertex objects.

    Notes
    -----
    This method takes a design and vertex ids, and gets their corresponding ``Vertex`` objects.
    """
    return [
        vertex
        for body in __traverse_all_bodies(design)
        for vertex in body.vertices
        if vertex.id in vertex_ids
    ]  # noqa: E501


def get_beams_from_ids(design: "Design", beam_ids: list[str]) -> list["Beam"]:
    """Find the ``Beam`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the beams.
    beam_ids : list[str]
        List of beam ids.

    Returns
    -------
    list[Beam]
        List of Beam objects.

    Notes
    -----
    This method takes a design and beam ids, and gets their corresponding ``Beam`` objects.
    """
    return [beam for beam in __traverse_all_beams(design) if beam.id in beam_ids]  # noqa: E501


def get_design_points_from_ids(
    design: "Design", design_point_ids: list[str]
) -> list["DesignPoint"]:
    """Find the ``DesignPoint`` objects inside a ``Design`` from its ids.

    Parameters
    ----------
    design : Design
        Parent design for the design points.
    design_point_ids : list[str]
        List of design point ids.

    Returns
    -------
    list[DesignPoint]
        List of DesignPoint objects.

    Notes
    -----
    This method takes a design and design point ids, and gets their corresponding ``DesignPoint``
    objects.
    """
    return [dp for dp in __traverse_all_design_points(design) if dp.id in design_point_ids]


def convert_color_to_hex(
    color: str | tuple[float, float, float] | tuple[float, float, float, float],
) -> str:
    """Get the hex string color from input formats.

    Parameters
    ----------
    color : str | tuple[float, float, float] | tuple[float, float, float, float]
        Color to set the body to. This can be a string representing a color name
        or a tuple of RGB values in the range [0, 1] (RGBA) or [0, 255] (pure RGB).

    Returns
    -------
    str
        The hex code string for the color, formatted #rrggbbaa.
    """
    import matplotlib.colors as mcolors

    try:
        if isinstance(color, tuple):
            # Ensure that all elements are within 0-1 or 0-255 range
            if all(0 <= c <= 1 for c in color):
                # Ensure they are floats if in 0-1 range
                if not all(isinstance(c, float) for c in color):
                    raise ValueError("RGB values in the 0-1 range must be floats.")
            elif all(0 <= c <= 255 for c in color):
                # Ensure they are integers if in 0-255 range
                if not all(isinstance(c, int) for c in color):
                    raise ValueError("RGB values in the 0-255 range must be integers.")
                # Normalize the 0-255 range to 0-1
                color = tuple(c / 255.0 for c in color)
            else:
                raise ValueError("RGB tuple contains mixed ranges or invalid values.")

            color = mcolors.to_hex(color, keep_alpha=True)
        elif isinstance(color, str):
            color = mcolors.to_hex(color, keep_alpha=True)
    except ValueError as err:
        raise ValueError(f"Invalid color value: {err}")

    return color


def convert_opacity_to_hex(opacity: float) -> str:
    """Get the hex string from an opacity value.

    Parameters
    ----------
    opacity : float
        Opacity to set body to. Must be in the range [0, 1].

    Returns
    -------
    The hex code for the opacity formatted #aa
    """
    try:
        # Ensure that the value is within 0-1 range
        if 0 <= opacity <= 1:
            return "{:02x}".format(int(opacity * 255))
        else:
            raise ValueError("Opacity value must be between 0 and 1.")
    except ValueError as err:
        raise ValueError(f"Invalid color value: {err}")


def prepare_file_for_server_upload(file_path: Path) -> Path:
    """Create a zip file from the given file path.

    Parameters
    ----------
    file_path : str
        The path to the file to be zipped.

    Returns
    -------
    Path
        The path to the created zip file.
    """
    import tempfile
    from zipfile import ZipFile

    # Create a temporary zip file with the same name as the original file
    temp_dir = Path(tempfile.gettempdir())
    temp_zip_path = temp_dir / f"{file_path.stem}.zip"

    # Create zip archive
    with ZipFile(temp_zip_path, "w") as zipf:
        # Add the main file
        zipf.write(file_path, file_path.name)

        # If it's an assembly format, add all files from the same directory
        assembly_extensions = [".CATProduct", ".asm", ".solution", ".sldasm"]
        if any(ext in str(file_path) for ext in assembly_extensions):
            dir_path = file_path.parent
            for file in dir_path.iterdir():
                if file.is_file() and file != file_path:
                    zipf.write(file, file.name)

    return temp_zip_path



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the PyAnsys Geometry ``geometry`` subpackage."""

from ansys.geometry.core.shapes.curves.circle import Circle, CircleEvaluation
from ansys.geometry.core.shapes.curves.curve import Curve
from ansys.geometry.core.shapes.curves.ellipse import Ellipse, EllipseEvaluation
from ansys.geometry.core.shapes.curves.line import Line, LineEvaluation
from ansys.geometry.core.shapes.curves.nurbs import NURBSCurve, NURBSCurveEvaluation
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
    ParamUV,
)
from ansys.geometry.core.shapes.surfaces.cone import Cone, ConeEvaluation
from ansys.geometry.core.shapes.surfaces.cylinder import Cylinder, CylinderEvaluation
from ansys.geometry.core.shapes.surfaces.plane import PlaneEvaluation, PlaneSurface
from ansys.geometry.core.shapes.surfaces.sphere import Sphere, SphereEvaluation
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.surface_evaluation import SurfaceEvaluation
from ansys.geometry.core.shapes.surfaces.torus import Torus
from ansys.geometry.core.shapes.surfaces.trimmed_surface import TrimmedSurface



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import grpc

from .._version import GeometryApiProtos, set_proto_version
from .base.admin import GRPCAdminService
from .base.assembly_condition import GRPCAssemblyConditionService
from .base.beams import GRPCBeamsService
from .base.bodies import GRPCBodyService
from .base.commands import GRPCCommandsService
from .base.commands_script import GRPCCommandsScriptService
from .base.components import GRPCComponentsService
from .base.coordinate_systems import GRPCCoordinateSystemService
from .base.curves import GRPCCurvesService
from .base.designs import GRPCDesignsService
from .base.driving_dimensions import GRPCDrivingDimensionsService
from .base.edges import GRPCEdgesService
from .base.faces import GRPCFacesService
from .base.materials import GRPCMaterialsService
from .base.measurement_tools import GRPCMeasurementToolsService
from .base.model_tools import GRPCModelToolsService
from .base.named_selection import GRPCNamedSelectionService
from .base.parts import GRPCPartsService
from .base.patterns import GRPCPatternsService
from .base.planes import GRPCPlanesService
from .base.points import GRPCPointsService
from .base.prepare_tools import GRPCPrepareToolsService
from .base.repair_tools import GRPCRepairToolsService
from .base.unsupported import GRPCUnsupportedService


class _GRPCServices:
    """
    Placeholder for the gRPC services (i.e. stubs).

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    version : GeometryApiProtos | str | None
        The version of the gRPC API protocol to use. If None, the latest
        version is used.

    Notes
    -----
    This class provides a unified interface to access the different
    gRPC services available in the Geometry API. It allows for easy
    switching between different versions of the API by using the
    `version` parameter in the constructor. The services are lazy-loaded
    to avoid unnecessary imports and to improve performance.
    """

    def __init__(self, channel: grpc.Channel, version: GeometryApiProtos | str | None = None):
        """
        Initialize the GRPCServices class.

        Parameters
        ----------
        channel : grpc.Channel
            The gRPC channel to the server.
        version : GeometryApiProtos | str | None
            The version of the gRPC API protocol to use. If None, the latest
            version is used.
        """
        # Set the proto version to be used
        self.version = set_proto_version(channel, version)
        self.channel = channel

        # Lazy load all the services
        self._admin = None
        self._assembly_condition = None
        self._beams = None
        self._bodies = None
        self._commands = None
        self._components = None
        self._coordinate_systems = None
        self._curves = None
        self._commands_script = None
        self._designs = None
        self._driving_dimensions = None
        self._edges = None
        self._faces = None
        self._materials = None
        self._measurement_tools = None
        self._model_tools = None
        self._named_selection = None
        self._parts = None
        self._patterns = None
        self._planes = None
        self._points = None
        self._prepare_tools = None
        self._repair_tools = None
        self._unsupported = None

    @property
    def admin(self) -> GRPCAdminService:
        """
        Get the admin service for the specified version.

        Returns
        -------
        GRPCAdminService
            The admin service for the specified version.
        """
        if not self._admin:
            # Import the appropriate admin service based on the version
            from .v0.admin import GRPCAdminServiceV0
            from .v1.admin import GRPCAdminServiceV1

            if self.version == GeometryApiProtos.V0:
                self._admin = GRPCAdminServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._admin = GRPCAdminServiceV1(self.channel)
            else:
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._admin

    @property
    def assembly_condition(self) -> GRPCAssemblyConditionService:
        """
        Get the assembly condition service for the specified version.

        Returns
        -------
        GRPCAssemblyConditionService
            The assembly condition service for the specified version.
        """
        if not self._assembly_condition:
            # Import the appropriate assembly condition service based on the version
            from .v0.assembly_condition import GRPCAssemblyConditionServiceV0
            from .v1.assembly_condition import GRPCAssemblyConditionServiceV1

            if self.version == GeometryApiProtos.V0:
                self._assembly_condition = GRPCAssemblyConditionServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._assembly_condition = GRPCAssemblyConditionServiceV1(self.channel)
            else:
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._assembly_condition

    @property
    def beams(self) -> GRPCBeamsService:
        """
        Get the Beams service for the specified version.

        Returns
        -------
        GRPCBeamsService
            The Beams service for the specified version.
        """
        if not self._beams:
            # Import the appropriate Beams service based on the version
            from .v0.beams import GRPCBeamsServiceV0
            from .v1.beams import GRPCBeamsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._beams = GRPCBeamsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._beams = GRPCBeamsServiceV1(self.channel)
            else:
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._beams

    @property
    def bodies(self) -> GRPCBodyService:
        """
        Get the body service for the specified version.

        Returns
        -------
        GRPCBodyService
            The body service for the specified version.
        """
        if not self._bodies:
            # Import the appropriate body service based on the version
            from .v0.bodies import GRPCBodyServiceV0
            from .v1.bodies import GRPCBodyServiceV1

            if self.version == GeometryApiProtos.V0:
                self._bodies = GRPCBodyServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._bodies = GRPCBodyServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._bodies

    @property
    def commands(self) -> GRPCCommandsService:
        """
        Get the commands service for the specified version.

        Returns
        -------
        GRPCCommandsService
            The commands service for the specified version.
        """
        if not self._commands:
            # Import the appropriate commands service based on the version
            from .v0.commands import GRPCCommandsServiceV0
            from .v1.commands import GRPCCommandsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._commands = GRPCCommandsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._commands = GRPCCommandsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._commands

    @property
    def components(self) -> GRPCComponentsService:
        """
        Get the components service for the specified version.

        Returns
        -------
        GRPCComponentsService
            The components service for the specified version.
        """
        if not self._components:
            # Import the appropriate components service based on the version
            from .v0.components import GRPCComponentsServiceV0
            from .v1.components import GRPCComponentsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._components = GRPCComponentsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._components = GRPCComponentsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._components

    @property
    def coordinate_systems(self) -> GRPCCoordinateSystemService:
        """
        Get the coordinate systems service for the specified version.

        Returns
        -------
        GRPCCoordinateSystemService
            The coordinate systems service for the specified version.
        """
        if not self._coordinate_systems:
            # Import the appropriate coordinate systems service based on the version
            from .v0.coordinate_systems import GRPCCoordinateSystemServiceV0
            from .v1.coordinate_systems import GRPCCoordinateSystemServiceV1

            if self.version == GeometryApiProtos.V0:
                self._coordinate_systems = GRPCCoordinateSystemServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._coordinate_systems = GRPCCoordinateSystemServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._coordinate_systems

    @property
    def curves(self) -> GRPCCurvesService:
        """
        Get the curves service for the specified version.

        Returns
        -------
        GRPCCurvesService
            The curves service for the specified version.
        """
        if not self._curves:
            # Import the appropriate curves service based on the version
            from .v0.curves import GRPCCurvesServiceV0
            from .v1.curves import GRPCCurvesServiceV1

            if self.version == GeometryApiProtos.V0:
                self._curves = GRPCCurvesServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._curves = GRPCCurvesServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._curves

    @property
    def designs(self) -> GRPCDesignsService:
        """
        Get the designs service for the specified version.

        Returns
        -------
        GRPCDesignsService
            The designs service for the specified version.
        """
        if not self._designs:
            # Import the appropriate designs service based on the version
            from .v0.designs import GRPCDesignsServiceV0
            from .v1.designs import GRPCDesignsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._designs = GRPCDesignsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._designs = GRPCDesignsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._designs

    @property
    def commands_script(self) -> GRPCCommandsScriptService:
        """
        Get the DBU application service for the specified version.

        Returns
        -------
        GRPCCommandsScriptService
            The commands script application service for the specified version.
        """
        if not self._commands_script:
            # Import the appropriate DBU application service based on the version
            from .v0.commands_script import GRPCCommandsScriptServiceV0
            from .v1.commands_script import GRPCCommandsScriptServiceV1

            if self.version == GeometryApiProtos.V0:
                self._commands_script = GRPCCommandsScriptServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._commands_script = GRPCCommandsScriptServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._commands_script

    @property
    def driving_dimensions(self) -> GRPCDrivingDimensionsService:
        """
        Get the driving dimensions service for the specified version.

        Returns
        -------
        GRPCDrivingDimensionsService
            The driving dimensions service for the specified version.
        """
        if not self._driving_dimensions:
            # Import the appropriate driving dimensions service based on the version
            from .v0.driving_dimensions import GRPCDrivingDimensionsServiceV0
            from .v1.driving_dimensions import GRPCDrivingDimensionsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._driving_dimensions = GRPCDrivingDimensionsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._driving_dimensions = GRPCDrivingDimensionsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._driving_dimensions

    @property
    def edges(self) -> GRPCEdgesService:
        """
        Get the edges service for the specified version.

        Returns
        -------
        GRPCEdgesService
            The edges service for the specified version.
        """
        if not self._edges:
            # Import the appropriate edges service based on the version
            from .v0.edges import GRPCEdgesServiceV0
            from .v1.edges import GRPCEdgesServiceV1

            if self.version == GeometryApiProtos.V0:
                self._edges = GRPCEdgesServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._edges = GRPCEdgesServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._edges

    @property
    def faces(self) -> GRPCFacesService:
        """
        Get the faces service for the specified version.

        Returns
        -------
        GRPCEdgesService
            The faces service for the specified version.
        """
        if not self._faces:
            # Import the appropriate faces service based on the version
            from .v0.faces import GRPCFacesServiceV0
            from .v1.faces import GRPCFacesServiceV1

            if self.version == GeometryApiProtos.V0:
                self._faces = GRPCFacesServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._faces = GRPCFacesServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._faces

    @property
    def materials(self) -> GRPCMaterialsService:
        """
        Get the materials service for the specified version.

        Returns
        -------
        GRPCMaterialsService
            The materials service for the specified version.
        """
        if not self._materials:
            # Import the appropriate materials service based on the version
            from .v0.materials import GRPCMaterialsServiceV0
            from .v1.materials import GRPCMaterialsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._materials = GRPCMaterialsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._materials = GRPCMaterialsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._materials

    @property
    def measurement_tools(self) -> GRPCMeasurementToolsService:
        """
        Get the measurement tools service for the specified version.

        Returns
        -------
        GRPCMeasurementToolsService
            The measurement tools service for the specified version.
        """
        if not self._measurement_tools:
            # Import the appropriate measurement tools service based on the version
            from .v0.measurement_tools import GRPCMeasurementToolsServiceV0
            from .v1.measurement_tools import GRPCMeasurementToolsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._measurement_tools = GRPCMeasurementToolsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._measurement_tools = GRPCMeasurementToolsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._measurement_tools

    @property
    def model_tools(self) -> GRPCModelToolsService:
        """
        Get the model tools service for the specified version.

        Returns
        -------
        GRPCModelToolsService
            The model tools service for the specified version.
        """
        if not self._model_tools:
            # Import the appropriate model tools service based on the version
            from .v0.model_tools import GRPCModelToolsServiceV0
            from .v1.model_tools import GRPCModelToolsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._model_tools = GRPCModelToolsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._model_tools = GRPCModelToolsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._model_tools

    @property
    def named_selection(self) -> GRPCNamedSelectionService:
        """
        Get the named selection service for the specified version.

        Returns
        -------
        GRPCNamedSelectionService
            The named selection service for the specified version.
        """
        if not self._named_selection:
            # Import the appropriate named selection service based on the version
            from .v0.named_selection import GRPCNamedSelectionServiceV0
            from .v1.named_selection import GRPCNamedSelectionServiceV1

            if self.version == GeometryApiProtos.V0:
                self._named_selection = GRPCNamedSelectionServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._named_selection = GRPCNamedSelectionServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._named_selection

    @property
    def parts(self) -> GRPCPartsService:
        """
        Get the parts service for the specified version.

        Returns
        -------
        GRPCPartsService
            The parts service for the specified version.
        """
        if not self._parts:
            # Import the appropriate parts service based on the version
            from .v0.parts import GRPCPartsServiceV0
            from .v1.parts import GRPCPartsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._parts = GRPCPartsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # The parts services is not required/made available in v1
                self._parts = GRPCPartsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._parts

    @property
    def patterns(self) -> GRPCPatternsService:
        """
        Get the patterns service for the specified version.

        Returns
        -------
        GRPCPatternsService
            The patterns service for the specified version.
        """
        if not self._patterns:
            # Import the appropriate patterns service based on the version
            from .v0.patterns import GRPCPatternsServiceV0
            from .v1.patterns import GRPCPatternsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._patterns = GRPCPatternsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._patterns = GRPCPatternsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._patterns

    @property
    def planes(self) -> GRPCPlanesService:
        """
        Get the planes service for the specified version.

        Returns
        -------
        GRPCPlanesService
            The planes service for the specified version.
        """
        if not self._planes:
            # Import the appropriate planes service based on the version
            from .v0.planes import GRPCPlanesServiceV0
            from .v1.planes import GRPCPlanesServiceV1

            if self.version == GeometryApiProtos.V0:
                self._planes = GRPCPlanesServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._planes = GRPCPlanesServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._planes

    @property
    def points(self) -> GRPCPointsService:
        """
        Get the points service for the specified version.

        Returns
        -------
        GRPCPointsService
            The points service for the specified version.
        """
        if not self._points:
            # Import the appropriate points service based on the version
            from .v0.points import GRPCPointsServiceV0
            from .v1.points import GRPCPointsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._points = GRPCPointsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._points = GRPCPointsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._points

    @property
    def prepare_tools(self) -> GRPCPrepareToolsService:
        """
        Get the prepare tools service for the specified version.

        Returns
        -------
        GRPCPrepareToolsService
            The prepare tools service for the specified version.
        """
        if not self._prepare_tools:
            # Import the appropriate prepare tools service based on the version
            from .v0.prepare_tools import GRPCPrepareToolsServiceV0
            from .v1.prepare_tools import GRPCPrepareToolsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._prepare_tools = GRPCPrepareToolsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._prepare_tools = GRPCPrepareToolsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._prepare_tools

    @property
    def repair_tools(self) -> GRPCRepairToolsService:
        """
        Get the repair tools service for the specified version.

        Returns
        -------
        RepairToolsServiceBase
            The repair tools service for the specified version.
        """
        if not self._repair_tools:
            from .v0.repair_tools import GRPCRepairToolsServiceV0
            from .v1.repair_tools import GRPCRepairToolsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._repair_tools = GRPCRepairToolsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._repair_tools = GRPCRepairToolsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._repair_tools

    @property
    def unsupported(self) -> GRPCUnsupportedService:
        """
        Get the unsupported service for the specified version.

        Returns
        -------
        GRPCUnsupportedService
            The unsupported service for the specified version.
        """
        if not self._unsupported:
            from .v0.unsupported import GRPCUnsupportedServiceV0
            from .v1.unsupported import GRPCUnsupportedServiceV1

            if self.version == GeometryApiProtos.V0:
                self._unsupported = GRPCUnsupportedServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:
                self._unsupported = GRPCUnsupportedServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._unsupported



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the Named Selection service implementation for v1."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.named_selection import GRPCNamedSelectionService
from .conversions import build_grpc_id


class GRPCNamedSelectionServiceV1(GRPCNamedSelectionService):
    """Named Selection service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    Named Selection service. It is specifically designed for the v1 version
    of the Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.discovery.v1.design.selections.namedselection_pb2_grpc import (
            NamedSelectionStub,
        )

        self.stub = NamedSelectionStub(channel)

    @protect_grpc
    def get_named_selection(self, **kwargs):  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import EntityRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = EntityRequest(id=build_grpc_id(kwargs["id"]))

        # Call the gRPC service
        response = self.stub.Get(request).named_selection

        # Return the response - formatted as a dictionary
        return {
            "id": response.id.id,
            "name": response.name,
            "bodies": [body.id.id for body in response.bodies],
            "faces": [face.id.id for face in response.faces],
            "edges": [edge.id.id for edge in response.edges],
            "beams": [beam.id.id for beam in response.beams],
            "design_points": [dp.id.id for dp in response.design_points],
            "components": [comp.id.id for comp in response.components],
            "vertices": [vertex.id.id for vertex in response.vertices],
        }

    @protect_grpc
    def create_named_selection(self, **kwargs):  # noqa: D102
        from ansys.api.discovery.v1.design.selections.namedselection_pb2 import (
            CreateRequest,
            CreateRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateRequest(
            request_data=[
                CreateRequestData(
                    name=kwargs["name"],
                    members_ids=[build_grpc_id(id) for id in kwargs["members"]],
                )
            ]
        )

        # Call the gRPC service
        response = self.stub.Create(request).named_selections[0]

        # Return the response - formatted as a dictionary
        return {
            "id": response.id.id,
            "name": response.name,
            "bodies": [body.id.id for body in response.bodies],
            "faces": [face.id.id for face in response.faces],
            "edges": [edge.id.id for edge in response.edges],
            "beams": [beam.id.id for beam in response.beams],
            "design_points": [dp.id.id for dp in response.design_points],
            "components": [comp.id.id for comp in response.components],
            "vertices": [vertex.id.id for vertex in response.vertices],
        }

    @protect_grpc
    def delete_named_selection(self, **kwargs):  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        self.stub.Delete(request)

        # Return the response - empty dictionary
        return {}

    @protect_grpc
    def rename_named_selection(self, **kwargs):  # noqa: D102
        from ansys.api.discovery.v1.design.designmessages_pb2 import (
            SetDesignEntityNameRequest,
            SetDesignEntityNameRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetDesignEntityNameRequest(
            request_data=[
                SetDesignEntityNameRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    name=kwargs["new_name"],
                )
            ]
        )

        # Call the gRPC service
        self.stub.SetName(request)

        # Return the response - empty dictionary
        return {}



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the faces service implementation for v0."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.conversions import (
    from_measurement_to_server_angle,
    from_measurement_to_server_length,
    to_area,
    to_distance,
)
from ..base.faces import GRPCFacesService
from .conversions import (
    build_grpc_id,
    from_grpc_curve_to_curve,
    from_grpc_direction_to_unit_vector,
    from_grpc_point_to_point3d,
    from_grpc_surface_to_surface,
    from_line_to_grpc_line,
    from_point3d_to_grpc_point,
    from_unit_vector_to_grpc_direction,
)


class GRPCFacesServiceV0(GRPCFacesService):
    """Faces service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    faces service. It is specifically designed for the v0 version of the
    Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
        from ansys.api.geometry.v0.faces_pb2_grpc import FacesStub

        self.stub = FacesStub(channel)
        self.commands_stub = CommandsStub(channel)

    @protect_grpc
    def get_surface(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetSurface(request=request)

        # Return the response - formatted as a dictionary
        return {
            "surface": from_grpc_surface_to_surface(response, kwargs["surface_type"]),
        }

    @protect_grpc
    def get_box_uv(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetBoxUV(request=request)

        # Return the response - formatted as a dictionary
        return {
            "uv_box": {
                "u": (response.start_u, response.end_u),
                "v": (response.start_v, response.end_v),
            }
        }

    @protect_grpc
    def get_area(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetArea(request=request)

        # Return the response - formatted as a dictionary
        return {"area": to_area(response.area)}

    @protect_grpc
    def get_edges(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetEdges(request=request)

        # Return the response - formatted as a dictionary
        return {
            "edges": [
                {
                    "id": edge.id,
                    "curve_type": edge.curve_type,
                    "is_reversed": edge.is_reversed,
                }
                for edge in response.edges
            ]
        }

    @protect_grpc
    def get_vertices(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        from ansys.api.geometry.v0.faces_pb2 import GetVerticesRequest

        request = GetVerticesRequest(face_id=build_grpc_id(kwargs["id"]))

        # Call the gRPC service
        response = self.stub.GetVertices(request=request)

        # Return the response - formatted as a dictionary
        return {
            "vertices": [
                {
                    "id": vertex.id.id,
                    "position": from_grpc_point_to_point3d(vertex.position),
                }
                for vertex in response.vertices
            ]
        }

    @protect_grpc
    def get_loops(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetLoops(request=request)

        # Return the response - formatted as a dictionary
        return {
            "loops": [
                {
                    "type": loop.type,
                    "length": to_distance(loop.length).value,
                    "min_corner": from_grpc_point_to_point3d(loop.bounding_box.min),
                    "max_corner": from_grpc_point_to_point3d(loop.bounding_box.max),
                    "edges": [edge for edge in loop.edges],
                }
                for loop in response.loops
            ]
        }

    @protect_grpc
    def get_color(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # Call the gRPC service
        response = self.stub.GetColor(request=request)

        # Return the response - formatted as a dictionary
        return {"color": response.color}

    @protect_grpc
    def get_bounding_box(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["id"])

        # If "tight" bounding box is requested, raise NotImplementedError as this is
        # not supported in v0
        if kwargs.get("tight", False):
            raise NotImplementedError(
                f"Method '{self.__class__.__name__}.get_bounding_box(..., tight=True)' is not "
                "implemented in this protofile version."
            )

        # Call the gRPC service
        response = self.stub.GetBoundingBox(request=request)

        # Return the response - formatted as a dictionary
        return {
            "min_corner": from_grpc_point_to_point3d(response.min),
            "max_corner": from_grpc_point_to_point3d(response.max),
            "center": from_grpc_point_to_point3d(response.center),
        }

    @protect_grpc
    def set_color(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2 import SetColorRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetColorRequest(
            face_id=kwargs["id"],
            color=kwargs["color"],
        )
        # Call the gRPC service
        response = self.stub.SetColor(request=request)

        # Return the response - formatted as a dictionary
        return {"success": response.success}

    @protect_grpc
    def get_normal(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2 import GetNormalRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetNormalRequest(
            id=kwargs["id"],
            u=kwargs["u"],
            v=kwargs["v"],
        )

        # Call the gRPC service
        response = self.stub.GetNormal(request=request)

        # Return the response - formatted as a dictionary
        return {
            "normal": from_grpc_direction_to_unit_vector(response.direction),
        }

    @protect_grpc
    def evaluate(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2 import EvaluateRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = EvaluateRequest(
            id=kwargs["id"],
            u=kwargs["u"],
            v=kwargs["v"],
        )

        # Call the gRPC service
        response = self.stub.Evaluate(request=request)

        # Return the response - formatted as a dictionary
        return {
            "point": from_grpc_point_to_point3d(response.point),
        }

    @protect_grpc
    def create_iso_parametric_curve(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2 import CreateIsoParamCurvesRequest

        from ansys.geometry.core.shapes.parameterization import Interval

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateIsoParamCurvesRequest(
            id=kwargs["id"],
            u_dir_curve=kwargs["use_u_param"],
            proportion=kwargs["parameter"],
        )

        # Call the gRPC service
        response = self.stub.CreateIsoParamCurves(request=request)

        # Return the response - formatted as a dictionary
        return {
            "curves": [
                {
                    "geometry": from_grpc_curve_to_curve(curve.curve),
                    "start": from_grpc_point_to_point3d(curve.start),
                    "end": from_grpc_point_to_point3d(curve.end),
                    "interval": Interval(curve.interval_start, curve.interval_end),
                    "length": to_distance(curve.length).value,
                }
                for curve in response.curves
            ]
        }

    @protect_grpc
    def extrude_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import ExtrudeFacesRequest

        # Assign direction
        direction = (
            None
            if kwargs["direction"] is None
            else from_unit_vector_to_grpc_direction(kwargs["direction"])
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = ExtrudeFacesRequest(
            faces=[build_grpc_id(id) for id in kwargs["face_ids"]],
            distance=from_measurement_to_server_length(kwargs["distance"]),
            direction=direction,
            extrude_type=kwargs["extrude_type"].value,
            pull_symmetric=kwargs["pull_symmetric"],
            offset_mode=kwargs["offset_mode"].value,
            copy=kwargs["copy"],
            force_do_as_extrude=kwargs["force_do_as_extrude"],
        )

        # Call the gRPC service
        response = self.commands_stub.ExtrudeFaces(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
        }

    @protect_grpc
    def extrude_faces_up_to(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import ExtrudeFacesUpToRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = ExtrudeFacesUpToRequest(
            faces=[build_grpc_id(id) for id in kwargs["face_ids"]],
            up_to_selection=build_grpc_id(kwargs["up_to_selection_id"]),
            seed_point=from_point3d_to_grpc_point(kwargs["seed_point"]),
            direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
            extrude_type=kwargs["extrude_type"].value,
            pull_symmetric=kwargs["pull_symmetric"],
            offset_mode=kwargs["offset_mode"].value,
            copy=kwargs["copy"],
            force_do_as_extrude=kwargs["force_do_as_extrude"],
        )

        # Call the gRPC service
        response = self.commands_stub.ExtrudeFacesUpTo(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
        }

    @protect_grpc
    def offset_faces_set_radius(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import OffsetFacesSetRadiusRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = OffsetFacesSetRadiusRequest(
            faces=[build_grpc_id(id) for id in kwargs["face_ids"]],
            radius=from_measurement_to_server_length(kwargs["radius"]),
            offset_mode=kwargs["offset_mode"].value,
            copy=kwargs["copy"],
            extrude_type=kwargs["extrude_type"].value,
        )

        # Call the gRPC service
        response = self.commands_stub.OffsetFacesSetRadius(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
        }

    @protect_grpc
    def revolve_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import RevolveFacesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = RevolveFacesRequest(
            selection=[build_grpc_id(id) for id in kwargs["selection_ids"]],
            axis=from_line_to_grpc_line(kwargs["axis"]),
            angle=from_measurement_to_server_angle(kwargs["angle"]),
            extrude_type=kwargs["extrude_type"].value,
        )

        # Call the gRPC service
        response = self.commands_stub.RevolveFaces(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
        }

    @protect_grpc
    def revolve_faces_up_to(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import RevolveFacesUpToRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = RevolveFacesUpToRequest(
            selection=[build_grpc_id(id) for id in kwargs["selection_ids"]],
            up_to_selection=build_grpc_id(kwargs["up_to_selection_id"]),
            axis=from_line_to_grpc_line(kwargs["axis"]),
            direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
            extrude_type=kwargs["extrude_type"].value,
        )

        # Call the gRPC service
        response = self.commands_stub.RevolveFacesUpTo(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
        }

    @protect_grpc
    def revolve_faces_by_helix(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import RevolveFacesByHelixRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = RevolveFacesByHelixRequest(
            selection=[build_grpc_id(id) for id in kwargs["selection_ids"]],
            axis=from_line_to_grpc_line(kwargs["axis"]),
            direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
            height=from_measurement_to_server_length(kwargs["height"]),
            pitch=from_measurement_to_server_length(kwargs["pitch"]),
            taper_angle=from_measurement_to_server_angle(kwargs["taper_angle"]),
            right_handed=kwargs["right_handed"],
            both_sides=kwargs["both_sides"],
            extrude_type=kwargs["extrude_type"].value,
        )

        # Call the gRPC service
        response = self.commands_stub.RevolveFacesByHelix(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
        }

    @protect_grpc
    def replace_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import ReplaceFaceRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = ReplaceFaceRequest(
            target_selection=[build_grpc_id(id) for id in kwargs["target_ids"]],
            replacement_selection=[build_grpc_id(id) for id in kwargs["replacement_ids"]],
        )

        # Call the gRPC service
        response = self.commands_stub.ReplaceFace(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
        }

    @protect_grpc
    def thicken_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import ThickenFacesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = ThickenFacesRequest(
            faces=[build_grpc_id(id) for id in kwargs["face_ids"]],
            direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
            value=from_measurement_to_server_length(kwargs["thickness"]),
            extrude_type=kwargs["extrude_type"].value,
            pull_symmetric=kwargs["pull_symmetric"],
            select_direction=kwargs["select_direction"],
        )

        # Call the gRPC service
        response = self.commands_stub.ThickenFaces(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
        }

    @protect_grpc
    def draft_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import DraftFacesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = DraftFacesRequest(
            faces=[build_grpc_id(id) for id in kwargs["face_ids"]],
            reference_faces=[build_grpc_id(id) for id in kwargs["reference_face_ids"]],
            draft_side=kwargs["draft_side"].value,
            draft_angle=from_measurement_to_server_angle(kwargs["angle"]),
            extrude_type=kwargs["extrude_type"].value,
        )

        # Call the gRPC server
        response = self.commands_stub.DraftFaces(request=request)

        # Return the drafted faces
        return {
            "created_faces": [face.id for face in response.created_faces],
        }

    @protect_grpc
    def get_round_info(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import RoundInfoRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = RoundInfoRequest(face=build_grpc_id(kwargs["face_id"]))

        # Call the gRPC service
        response = self.commands_stub.GetRoundInfo(request=request)

        # Return the response - formatted as a dictionary
        return {
            "along_u": response.along_u,
            "radius": response.radius,
        }

    @protect_grpc
    def offset_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.faces_pb2 import OffsetFacesRequest, OffsetFacesRequestData

        # Create the request - assumes all inputs are valid and of the proper type
        request = OffsetFacesRequest(
            request_data=[
                OffsetFacesRequestData(
                    faces=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    offset=from_measurement_to_server_length(kwargs["distance"]),
                    direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
                    extrude_type=kwargs["extrude_type"].value,
                )
            ]
        )

        # Call the gRPC service
        response = self.stub.OffsetFaces(request=request)

        # Return the response - formatted as a dictionary
        return {
            "results": [
                [face.id for face in response_data.new_faces]
                for response_data in response.response_data
            ]
        }

    @protect_grpc
    def setup_offset_relationship(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import FaceOffsetRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FaceOffsetRequest(
            face1=build_grpc_id(kwargs["face1_id"]),
            face2=build_grpc_id(kwargs["face2_id"]),
            set_baselines=kwargs["set_baselines"],
            process_adjacent_faces=kwargs["process_adjacent_faces"],
        )

        # Call the gRPC service
        response = self.commands_stub.FaceOffset(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
        }



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides tools for pulling geometry."""

from enum import Enum, unique
from typing import TYPE_CHECKING, Union

from beartype import beartype as check_input_types
from pint import Quantity

import ansys.geometry.core as pyansys_geo
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.designer.component import Component
from ansys.geometry.core.designer.mating_conditions import (
    AlignCondition,
    OrientCondition,
    TangentCondition,
)
from ansys.geometry.core.designer.selection import NamedSelection
from ansys.geometry.core.errors import GeometryRuntimeError
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.auxiliary import (
    get_bodies_from_ids,
    get_design_from_body,
    get_design_from_component,
    get_design_from_edge,
    get_design_from_face,
    get_faces_from_ids,
)
from ansys.geometry.core.misc.checks import (
    check_type,
    check_type_all_elements_in_iterable,
    min_backend_version,
)
from ansys.geometry.core.misc.measurements import Angle, Distance
from ansys.geometry.core.shapes.curves.curve import Curve
from ansys.geometry.core.shapes.curves.line import Line
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.component import Component
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face


@unique
class ExtrudeType(Enum):
    """Provides values for extrusion types."""

    NONE = 0
    ADD = 1
    CUT = 2
    FORCE_ADD = 3
    FORCE_CUT = 4
    FORCE_INDEPENDENT = 5
    FORCE_NEW_SURFACE = 6


@unique
class OffsetMode(Enum):
    """Provides values for offset modes during extrusions."""

    IGNORE_RELATIONSHIPS = 0
    MOVE_FACES_TOGETHER = 1
    MOVE_FACES_APART = 2


@unique
class FillPatternType(Enum):
    """Provides values for types of fill patterns."""

    GRID = 0
    OFFSET = 1
    SKEWED = 2


@unique
class DraftSide(Enum):
    """Provides values for draft sides."""

    NO_SPLIT = 0
    THIS = 1
    OTHER = 2
    BACK = 3


class GeometryCommands:
    """Provides geometry commands for PyAnsys Geometry.

    Parameters
    ----------
    grpc_client : GrpcClient
        gRPC client to use for the geometry commands.
    _internal_use : bool, optional
        Internal flag to prevent direct instantiation by users.
        This parameter is for internal use only.

    Raises
    ------
    GeometryRuntimeError
        If the class is instantiated directly by users instead
        of through the modeler.

    Notes
    -----
    This class should not be instantiated directly. Use
    ``modeler.geometry_commands`` instead.
    """

    def __init__(self, grpc_client: GrpcClient, _internal_use: bool = False):
        """Initialize an instance of the ``GeometryCommands`` class."""
        if not _internal_use:
            raise GeometryRuntimeError(
                "GeometryCommands should not be instantiated directly. "
                "Use 'modeler.geometry_commands' to access geometry commands."
            )
        self._grpc_client = grpc_client

    @min_backend_version(25, 2, 0)
    def chamfer(
        self,
        selection: Union["Edge", list["Edge"], "Face", list["Face"]],
        distance: Distance | Quantity | Real,
    ) -> bool:
        """Create a chamfer on an edge or adjust the chamfer of a face.

        Parameters
        ----------
        selection : Edge | list[Edge] | Face | list[Face]
            One or more edges or faces to act on.
        distance : Distance | Quantity | Real
            Chamfer distance.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.edge import Edge
        from ansys.geometry.core.designer.face import Face

        selection: list[Edge | Face] = selection if isinstance(selection, list) else [selection]

        check_type_all_elements_in_iterable(selection, (Edge, Face))

        # Convert the distance object
        distance = distance if isinstance(distance, Distance) else Distance(distance)

        for ef in selection:
            ef.body._reset_tessellation_cache()

        result = self._grpc_client.services.model_tools.chamfer(
            selection_ids=[ef.id for ef in selection],
            distance=distance,
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def fillet(
        self,
        selection: Union["Edge", list["Edge"], "Face", list["Face"]],
        radius: Distance | Quantity | Real,
    ) -> bool:
        """Create a fillet on an edge or adjust the fillet of a face.

        Parameters
        ----------
        selection : Edge | list[Edge] | Face | list[Face]
            One or more edges or faces to act on.
        radius : Distance | Quantity | Real
            Fillet radius.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.edge import Edge
        from ansys.geometry.core.designer.face import Face

        selection: list[Edge | Face] = selection if isinstance(selection, list) else [selection]

        check_type_all_elements_in_iterable(selection, (Edge, Face))

        # Convert the radius object
        radius = radius if isinstance(radius, Distance) else Distance(radius)

        for ef in selection:
            ef.body._reset_tessellation_cache()

        result = self._grpc_client.services.model_tools.fillet(
            selection_ids=[ef.id for ef in selection],
            radius=radius,
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def full_fillet(self, faces: list["Face"]) -> bool:
        """Create a full fillet betweens a collection of faces.

        Parameters
        ----------
        faces : list[Face]
            Faces to round.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        check_type_all_elements_in_iterable(faces, Face)

        for face in faces:
            face.body._reset_tessellation_cache()

        result = self._grpc_client.services.model_tools.full_fillet(
            selection_ids=[face.id for face in faces],
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def extrude_faces(
        self,
        faces: Union["Face", list["Face"]],
        distance: Distance | Quantity | Real,
        direction: UnitVector3D = None,
        extrude_type: ExtrudeType = ExtrudeType.ADD,
        offset_mode: OffsetMode = OffsetMode.MOVE_FACES_TOGETHER,
        pull_symmetric: bool = False,
        copy: bool = False,
        force_do_as_extrude: bool = False,
    ) -> list["Body"]:
        """Extrude a selection of faces.

        Parameters
        ----------
        faces : Face | list[Face]
            Faces to extrude.
        distance : Distance | Quantity | Real
            Distance to extrude.
        direction : UnitVector3D, default: None
            Direction of extrusion. If no direction is provided, it will be inferred.
        extrude_type : ExtrudeType, default: ExtrudeType.ADD
            Type of extrusion to be performed.
        offset_mode : OffsetMode, default: OffsetMode.MOVE_FACES_TOGETHER
            Mode of how to handle offset relationships.
        pull_symmetric : bool, default: False
            Pull symmetrically on both sides if ``True``.
        copy : bool, default: False
            Copy the face and move it instead of extruding the original face if ``True``.
        force_do_as_extrude : bool, default: False
            Forces to do as an extrusion if ``True``, if ``False`` allows extrusion by offset.

        Returns
        -------
        list[Body]
            Bodies created by the extrusion if any.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        faces: list[Face] = faces if isinstance(faces, list) else [faces]
        check_type_all_elements_in_iterable(faces, Face)

        # Create distance object
        distance = distance if isinstance(distance, Distance) else Distance(distance)

        for face in faces:
            face.body._reset_tessellation_cache()

        result = self._grpc_client.services.faces.extrude_faces(
            face_ids=[face.id for face in faces],
            distance=distance,
            direction=direction,
            extrude_type=extrude_type,
            pull_symmetric=pull_symmetric,
            offset_mode=offset_mode,
            copy=copy,
            force_do_as_extrude=force_do_as_extrude,
        )

        design = get_design_from_face(faces[0])

        if result.get("success"):
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                design._update_from_tracker(result.get("tracked_response"))
            else:
                design._update_design_inplace()
            return get_bodies_from_ids(design, result.get("created_bodies"))
        else:
            self._grpc_client.log.info("Failed to extrude faces.")
            return []

    @min_backend_version(25, 2, 0)
    def extrude_faces_up_to(
        self,
        faces: Union["Face", list["Face"]],
        up_to_selection: Union["Face", "Edge", "Body"],
        seed_point: Point3D,
        direction: UnitVector3D,
        extrude_type: ExtrudeType = ExtrudeType.ADD,
        offset_mode: OffsetMode = OffsetMode.MOVE_FACES_TOGETHER,
        pull_symmetric: bool = False,
        copy: bool = False,
        force_do_as_extrude: bool = False,
    ) -> list["Body"]:
        """Extrude a selection of faces up to another object.

        Parameters
        ----------
        faces : Face | list[Face]
            Faces to extrude.
        up_to_selection : Face | Edge | Body
            The object to pull the faces up to.
        seed_point : Point3D
            Origin to define the extrusion.
        direction : UnitVector3D, default: None
            Direction of extrusion. If no direction is provided, it will be inferred.
        extrude_type : ExtrudeType, default: ExtrudeType.ADD
            Type of extrusion to be performed.
        offset_mode : OffsetMode, default: OffsetMode.MOVE_FACES_TOGETHER
            Mode of how to handle offset relationships.
        pull_symmetric : bool, default: False
            Pull symmetrically on both sides if ``True``.
        copy : bool, default: False
            Copy the face and move it instead of extruding the original face if ``True``.
        force_do_as_extrude : bool, default: False
            Forces to do as an extrusion if ``True``, if ``False`` allows extrusion by offset.

        Returns
        -------
        list[Body]
            Bodies created by the extrusion if any.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        faces: list[Face] = faces if isinstance(faces, list) else [faces]
        check_type_all_elements_in_iterable(faces, Face)

        for face in faces:
            face.body._reset_tessellation_cache()

        result = self._grpc_client.services.faces.extrude_faces_up_to(
            face_ids=[face.id for face in faces],
            up_to_selection_id=up_to_selection.id,
            seed_point=seed_point,
            direction=direction,
            extrude_type=extrude_type,
            pull_symmetric=pull_symmetric,
            offset_mode=offset_mode,
            copy=copy,
            force_do_as_extrude=force_do_as_extrude,
        )

        design = get_design_from_face(faces[0])

        if result.get("success"):
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                design._update_from_tracker(result.get("tracked_response"))
            else:
                design._update_design_inplace()
            return get_bodies_from_ids(design, result.get("created_bodies"))
        else:
            self._grpc_client.log.info("Failed to extrude faces.")
            return []

    @min_backend_version(25, 2, 0)
    def extrude_edges(
        self,
        edges: Union["Edge", list["Edge"]],
        distance: Distance | Quantity | Real,
        from_face: "Face" = None,
        from_point: Point3D = None,
        direction: UnitVector3D = None,
        extrude_type: ExtrudeType = ExtrudeType.ADD,
        pull_symmetric: bool = False,
        copy: bool = False,
        natural_extension: bool = False,
    ) -> list["Body"]:
        """Extrude a selection of edges. Provide either a face or a direction and point.

        Parameters
        ----------
        edges : Edge | list[Edge]
            Edges to extrude.
        distance : Distance | Quantity | Real
            Distance to extrude.
        from_face : Face, default: None
            Face to pull normal from.
        from_point : Point3D, default: None
            Point to pull from. Must be used with ``direction``.
        direction : UnitVector3D, default: None
            Direction to pull. Must be used with ``from_point``.
        extrude_type : ExtrudeType, default: ExtrudeType.ADD
            Type of extrusion to be performed.
        pull_symmetric : bool, default: False
            Pull symmetrically on both sides if ``True``.
        copy : bool, default: False
            Copy the edge and move it instead of extruding the original edge if ``True``.
        natural_extension : bool, default: False
            Surfaces will extend in a natural or linear shape after exceeding its original range.

        Returns
        -------
        list[Body]
            Bodies created by the extrusion if any.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.edge import Edge

        edges: list[Edge] = edges if isinstance(edges, list) else [edges]
        check_type_all_elements_in_iterable(edges, Edge)

        # Create distance object
        distance = distance if isinstance(distance, Distance) else Distance(distance)

        if from_face is None and None in (from_point, direction):
            raise ValueError(
                "To extrude edges, either a face or a direction and point must be provided."
            )

        for edge in edges:
            edge.body._reset_tessellation_cache()

        result = self._grpc_client.services.edges.extrude_edges(
            edge_ids=[edge.id for edge in edges],
            distance=distance,
            face=from_face.id,
            point=from_point,
            direction=direction,
            extrude_type=extrude_type,
            pull_symmetric=pull_symmetric,
            copy=copy,
            natural_extension=natural_extension,
        )

        design = get_design_from_edge(edges[0])

        if result.get("success"):
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                design._update_from_tracker(result.get("tracked_response"))
            else:
                design._update_design_inplace()
            return get_bodies_from_ids(design, result.get("created_bodies"))
        else:
            self._grpc_client.log.info("Failed to extrude edges.")
            return []

    @min_backend_version(25, 2, 0)
    def extrude_edges_up_to(
        self,
        edges: Union["Edge", list["Edge"]],
        up_to_selection: Union["Face", "Edge", "Body"],
        seed_point: Point3D,
        direction: UnitVector3D,
        extrude_type: ExtrudeType = ExtrudeType.ADD,
    ) -> list["Body"]:
        """Extrude a selection of edges up to another object.

        Parameters
        ----------
        edges : Edge | list[Edge]
            Edges to extrude.
        up_to_selection : Face, default: None
            The object to pull the faces up to.
        seed_point : Point3D
            Origin to define the extrusion.
        direction : UnitVector3D, default: None
            Direction of extrusion.
        extrude_type : ExtrudeType, default: ExtrudeType.ADD
            Type of extrusion to be performed.

        Returns
        -------
        list[Body]
            Bodies created by the extrusion if any.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.edge import Edge

        edges: list[Edge] = edges if isinstance(edges, list) else [edges]
        check_type_all_elements_in_iterable(edges, Edge)

        for edge in edges:
            edge.body._reset_tessellation_cache()

        result = self._grpc_client.services.edges.extrude_edges_up_to(
            edge_ids=[edge.id for edge in edges],
            up_to_selection=up_to_selection.id,
            seed_point=seed_point,
            direction=direction,
            extrude_type=extrude_type,
        )

        design = get_design_from_edge(edges[0])

        if result.get("success"):
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                design._update_from_tracker(result.get("tracked_response"))
            else:
                design._update_design_inplace()
            return get_bodies_from_ids(design, result.get("created_bodies"))
        else:
            self._grpc_client.log.info("Failed to extrude edges.")
            return []

    @min_backend_version(25, 2, 0)
    def rename_object(
        self,
        selection: Union[list["Body"], list["Component"]],
        name: str,
    ) -> bool:
        """Rename an object.

        Parameters
        ----------
        selection : list[Body] | list[Component]
            Selection of the objects to rename.
        name : str
            New name for the object.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        result = self._grpc_client._services.commands.set_name(
            selection_ids=[object.id for object in selection], name=name
        )
        return result.get("success")

    @min_backend_version(25, 2, 0)
    def create_linear_pattern(
        self,
        selection: Union["Face", list["Face"]],
        linear_direction: Union["Edge", "Face"],
        count_x: int,
        pitch_x: Distance | Quantity | Real,
        two_dimensional: bool = False,
        count_y: int = None,
        pitch_y: Distance | Quantity | Real = None,
    ) -> bool:
        """Create a linear pattern. The pattern can be one or two dimensions.

        Parameters
        ----------
        selection : Face | list[Face]
            Faces to create the pattern out of.
        linear_direction : Edge | Face
            Direction of the linear pattern, determined by the direction of an edge or face normal.
        count_x : int
            How many times the pattern repeats in the x direction.
        pitch_x : Distance | Quantity | Real
            The spacing between each pattern member in the x direction.
        two_dimensional : bool, default: False
            If ``True``, create a pattern in the x and y direction.
        count_y : int, default: None
            How many times the pattern repeats in the y direction.
        pitch_y : Distance | Quantity | Real, default: None
            The spacing between each pattern member in the y direction.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        selection: list[Face] = selection if isinstance(selection, list) else [selection]

        check_type_all_elements_in_iterable(selection, Face)

        for object in selection:
            object.body._reset_tessellation_cache()

        if two_dimensional and None in (count_y, pitch_y):
            raise ValueError(
                "If the pattern is two dimensional, count_y and pitch_y must be provided."
            )
        if not two_dimensional and None not in (count_y, pitch_y):
            raise ValueError(
                (
                    "You provided count_y and pitch_y. Ensure two_dimensional is True if a "
                    "two-dimensional pattern is desired."
                )
            )

        # Convert pitches to distance objects
        pitch_x = pitch_x if isinstance(pitch_x, Distance) else Distance(pitch_x)
        if pitch_y is not None:
            pitch_y = pitch_y if isinstance(pitch_y, Distance) else Distance(pitch_y)
        else:
            pitch_y = Distance(0)

        result = self._grpc_client.services.patterns.create_linear_pattern(
            selection_ids=[object.id for object in selection],
            linear_direction_id=linear_direction.id,
            count_x=count_x,
            pitch_x=pitch_x,
            two_dimensional=two_dimensional,
            count_y=count_y,
            pitch_y=pitch_y,
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def modify_linear_pattern(
        self,
        selection: Union["Face", list["Face"]],
        count_x: int = 0,
        pitch_x: Distance | Quantity | Real = 0.0,
        count_y: int = 0,
        pitch_y: Distance | Quantity | Real = 0.0,
        new_seed_index: int = 0,
        old_seed_index: int = 0,
    ) -> bool:
        """Modify a linear pattern. Leave an argument at 0 for it to remain unchanged.

        Parameters
        ----------
        selection : Face | list[Face]
            Faces that belong to the pattern.
        count_x : int, default: 0
            How many times the pattern repeats in the x direction.
        pitch_x : Distance | Quantity | Real, default: 0.0
            The spacing between each pattern member in the x direction.
        count_y : int, default: 0
            How many times the pattern repeats in the y direction.
        pitch_y : Distance | Quantity | Real, default: 0.0
            The spacing between each pattern member in the y direction.
        new_seed_index : int, default: 0
            The new seed index of the member.
        old_seed_index : int, default: 0
            The old seed index of the member.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        selection: list[Face] = selection if isinstance(selection, list) else [selection]

        check_type_all_elements_in_iterable(selection, Face)

        for object in selection:
            object.body._reset_tessellation_cache()

        # Convert pitches to distance objects
        pitch_x = pitch_x if isinstance(pitch_x, Distance) else Distance(pitch_x)
        pitch_y = pitch_y if isinstance(pitch_y, Distance) else Distance(pitch_y)

        result = self._grpc_client.services.patterns.modify_linear_pattern(
            selection_ids=[object.id for object in selection],
            count_x=count_x,
            pitch_x=pitch_x,
            count_y=count_y,
            pitch_y=pitch_y,
            new_seed_index=new_seed_index,
            old_seed_index=old_seed_index,
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def create_circular_pattern(
        self,
        selection: Union["Face", list["Face"]],
        circular_axis: Union["Edge", Line],
        circular_count: int,
        circular_angle: Angle | Quantity | Real,
        two_dimensional: bool = False,
        linear_count: int = None,
        linear_pitch: Distance | Quantity | Real = None,
        radial_direction: UnitVector3D = None,
    ) -> bool:
        """Create a circular pattern. The pattern can be one or two dimensions.

        Parameters
        ----------
        selection : Face | list[Face]
            Faces to create the pattern out of.
        circular_axis : Edge | Line
            The axis of the circular pattern.
        circular_count : int
            How many members are in the circular pattern.
        circular_angle : Angle | Quantity | Real
            The angular range of the pattern.
        two_dimensional : bool, default: False
            If ``True``, create a two-dimensional pattern.
        linear_count : int, default: None
            How many times the circular pattern repeats along the radial lines for a
            two-dimensional pattern.
        linear_pitch : Distance | Quantity | Real, default: None
            The spacing along the radial lines for a two-dimensional pattern.
        radial_direction : UnitVector3D, default: None
            The direction from the center out for a two-dimensional pattern.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.edge import Edge
        from ansys.geometry.core.designer.face import Face

        selection: list[Face] = selection if isinstance(selection, list) else [selection]

        check_type_all_elements_in_iterable(selection, Face)

        for object in selection:
            object.body._reset_tessellation_cache()

        if two_dimensional and None in (linear_count, linear_pitch):
            raise ValueError(
                "If the pattern is two-dimensional, linear_count and linear_pitch must be provided."
            )
        if not two_dimensional and None not in (
            linear_count,
            linear_pitch,
        ):
            raise ValueError(
                (
                    "You provided linear_count and linear_pitch. Ensure two_dimensional is True if "
                    "a two-dimensional pattern is desired."
                )
            )
        if self._grpc_client.backend_version < (26, 1, 0) and isinstance(circular_axis, Line):
            raise ValueError(
                (
                    "Using a Line as the circular axis is only supported "
                    "starting in Ansys release 26R1."
                )
            )

        # Convert angle and pitch to appropriate objects
        if not isinstance(circular_angle, Angle):
            circular_angle = Angle(circular_angle)
        if linear_pitch is not None:
            linear_pitch = (
                linear_pitch if isinstance(linear_pitch, Distance) else Distance(linear_pitch)
            )
        else:
            linear_pitch = Distance(0)
        if isinstance(circular_axis, Edge):
            circular_axis = circular_axis.id

        result = self._grpc_client.services.patterns.create_circular_pattern(
            selection_ids=[object.id for object in selection],
            circular_axis=circular_axis,
            circular_count=circular_count,
            circular_angle=circular_angle,
            two_dimensional=two_dimensional,
            linear_count=linear_count,
            linear_pitch=linear_pitch,
            radial_direction=radial_direction,
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def modify_circular_pattern(
        self,
        selection: Union["Face", list["Face"]],
        circular_count: int = 0,
        linear_count: int = 0,
        step_angle: Angle | Quantity | Real = 0.0,
        step_linear: Distance | Quantity | Real = 0.0,
    ) -> bool:
        """Modify a circular pattern. Leave an argument at 0 for it to remain unchanged.

        Parameters
        ----------
        selection : Face | list[Face]
            Faces that belong to the pattern.
        circular_count : int, default: 0
            How many members are in the circular pattern.
        linear_count : int, default: 0
            How many times the circular pattern repeats along the radial lines for a
            two-dimensional pattern.
        step_angle : Angle | Quantity | Real, default: 0.0
            Defines the circular angle.
        step_linear : Distance | Quantity | Real, default: 0.0
            Defines the step, along the radial lines, for a pattern dimension greater than 1.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        selection: list[Face] = selection if isinstance(selection, list) else [selection]

        check_type_all_elements_in_iterable(selection, Face)

        for object in selection:
            object.body._reset_tessellation_cache()

        # Convert angle and pitch to appropriate objects
        step_angle = step_angle if isinstance(step_angle, Angle) else Angle(step_angle)
        step_linear = step_linear if isinstance(step_linear, Distance) else Distance(step_linear)

        result = self._grpc_client.services.patterns.modify_circular_pattern(
            selection_ids=[object.id for object in selection],
            circular_count=circular_count,
            linear_count=linear_count,
            step_angle=step_angle,
            step_linear=step_linear,
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def create_fill_pattern(
        self,
        selection: Union["Face", list["Face"]],
        linear_direction: Union["Edge", "Face"],
        fill_pattern_type: FillPatternType,
        margin: Distance | Quantity | Real,
        x_spacing: Distance | Quantity | Real,
        y_spacing: Distance | Quantity | Real,
        row_x_offset: Distance | Quantity | Real = 0,
        row_y_offset: Distance | Quantity | Real = 0,
        column_x_offset: Distance | Quantity | Real = 0,
        column_y_offset: Distance | Quantity | Real = 0,
    ) -> bool:
        """Create a fill pattern.

        Parameters
        ----------
        selection : Face | list[Face]
            Faces to create the pattern out of.
        linear_direction : Edge
            Direction of the linear pattern, determined by the direction of an edge.
        fill_pattern_type : FillPatternType
            The type of fill pattern.
        margin : Distance | Quantity | Real
            Margin defining the border of the fill pattern.
        x_spacing : Distance | Quantity | Real
            Spacing between the pattern members in the x direction.
        y_spacing : Distance | Quantity | Real
            Spacing between the pattern members in the x direction.
        row_x_offset : Distance | Quantity | Real, default: 0
            Offset for the rows in the x direction. Only used with ``FillPattern.SKEWED``.
        row_y_offset : Distance | Quantity | Real, default: 0
            Offset for the rows in the y direction. Only used with ``FillPattern.SKEWED``.
        column_x_offset : Distance | Quantity | Real, default: 0
            Offset for the columns in the x direction. Only used with ``FillPattern.SKEWED``.
        column_y_offset : Distance | Quantity | Real, default: 0
            Offset for the columns in the y direction. Only used with ``FillPattern.SKEWED``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        selection: list[Face] = selection if isinstance(selection, list) else [selection]

        check_type_all_elements_in_iterable(selection, Face)

        for object in selection:
            object.body._reset_tessellation_cache()

        # Convert measurements to distance objects
        margin = margin if isinstance(margin, Distance) else Distance(margin)
        x_spacing = x_spacing if isinstance(x_spacing, Distance) else Distance(x_spacing)
        y_spacing = y_spacing if isinstance(y_spacing, Distance) else Distance(y_spacing)
        row_x_offset = (
            row_x_offset if isinstance(row_x_offset, Distance) else Distance(row_x_offset)
        )
        row_y_offset = (
            row_y_offset if isinstance(row_y_offset, Distance) else Distance(row_y_offset)
        )
        column_x_offset = (
            column_x_offset if isinstance(column_x_offset, Distance) else Distance(column_x_offset)
        )
        column_y_offset = (
            column_y_offset if isinstance(column_y_offset, Distance) else Distance(column_y_offset)
        )

        result = self._grpc_client.services.patterns.create_fill_pattern(
            selection_ids=[object.id for object in selection],
            linear_direction_id=linear_direction.id,
            fill_pattern_type=fill_pattern_type,
            margin=margin,
            x_spacing=x_spacing,
            y_spacing=y_spacing,
            row_x_offset=row_x_offset,
            row_y_offset=row_y_offset,
            column_x_offset=column_x_offset,
            column_y_offset=column_y_offset,
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def update_fill_pattern(
        self,
        selection: Union["Face", list["Face"]],
    ) -> bool:
        """Update a fill pattern.

        When the face that a fill pattern exists upon changes in size, the
        fill pattern can be updated to fill the new space.

        Parameters
        ----------
        selection : Face | list[Face]
            Face(s) that are part of a fill pattern.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        selection: list[Face] = selection if isinstance(selection, list) else [selection]

        check_type_all_elements_in_iterable(selection, Face)

        for object in selection:
            object.body._reset_tessellation_cache()

        result = self._grpc_client.services.patterns.update_fill_pattern(
            selection_ids=[object.id for object in selection],
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def revolve_faces(
        self,
        selection: Union["Face", list["Face"]],
        axis: Union["Edge", Line],
        angle: Angle | Quantity | Real,
        extrude_type: ExtrudeType = ExtrudeType.ADD,
    ) -> list["Body"]:
        """Revolve face around an axis.

        Parameters
        ----------
        selection : Face | list[Face]
            Face(s) to revolve.
        axis : Edge | Line
            Axis of revolution.
        angle : Angle | Quantity | Real
            Angular distance to revolve.
        extrude_type : ExtrudeType, default: ExtrudeType.ADD
            Type of extrusion to be performed.

        Returns
        -------
        list[Body]
            Bodies created by the extrusion if any.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.edge import CurveType, Edge
        from ansys.geometry.core.designer.face import Face

        selection: list[Face] = selection if isinstance(selection, list) else [selection]
        check_type_all_elements_in_iterable(selection, Face)

        angle = angle if isinstance(angle, Angle) else Angle(angle)

        for object in selection:
            object.body._reset_tessellation_cache()

        if isinstance(axis, Edge):
            if axis.curve_type != CurveType.CURVETYPE_LINE:
                raise ValueError("Only edges that are lines can be used as the revolve axis.")
            axis = Line(axis.start, UnitVector3D.from_points(axis.start, axis.end))

        result = self._grpc_client._services.faces.revolve_faces(
            selection_ids=[object.id for object in selection],
            axis=axis,
            angle=angle,
            extrude_type=extrude_type,
        )

        design = get_design_from_face(selection[0])

        if result.get("success"):
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                design._update_from_tracker(result.get("tracked_response"))
            else:
                design._update_design_inplace()
            return get_bodies_from_ids(design, result.get("created_bodies"))
        else:
            self._grpc_client.log.info("Failed to revolve faces.")
            return []

    @min_backend_version(25, 2, 0)
    def revolve_faces_up_to(
        self,
        selection: Union["Face", list["Face"]],
        up_to: Union["Face", "Edge", "Body"],
        axis: Line,
        direction: UnitVector3D,
        extrude_type: ExtrudeType = ExtrudeType.ADD,
    ) -> list["Body"]:
        """Revolve face around an axis up to a certain object.

        Parameters
        ----------
        selection : Face | list[Face]
            Face(s) to revolve.
        up_to : Face | Edge | Body
            Object to revolve the face up to.
        axis : Line
            Axis of revolution.
        direction : UnitVector3D
            Direction of extrusion.
        extrude_type : ExtrudeType, default: ExtrudeType.ADD
            Type of extrusion to be performed.

        Returns
        -------
        list[Body]
            Bodies created by the extrusion if any.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        selection: list[Face] = selection if isinstance(selection, list) else [selection]
        check_type_all_elements_in_iterable(selection, Face)

        for object in selection:
            object.body._reset_tessellation_cache()

        result = self._grpc_client._services.faces.revolve_faces_up_to(
            selection_ids=[object.id for object in selection],
            up_to_selection_id=up_to.id,
            axis=axis,
            direction=direction,
            extrude_type=extrude_type,
        )

        design = get_design_from_face(selection[0])

        if result.get("success"):
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                design._update_from_tracker(result.get("tracked_response"))
            else:
                design._update_design_inplace()
            return get_bodies_from_ids(design, result.get("created_bodies"))
        else:
            self._grpc_client.log.info("Failed to revolve faces.")
            return []

    @min_backend_version(25, 2, 0)
    def revolve_faces_by_helix(
        self,
        selection: Union["Face", list["Face"]],
        axis: Line,
        direction: UnitVector3D,
        height: Distance | Quantity | Real,
        pitch: Distance | Quantity | Real,
        taper_angle: Angle | Quantity | Real,
        right_handed: bool,
        both_sides: bool,
        extrude_type: ExtrudeType = ExtrudeType.ADD,
    ) -> list["Body"]:
        """Revolve face around an axis in a helix shape.

        Parameters
        ----------
        selection : Face | list[Face]
            Face(s) to revolve.
        axis : Line
            Axis of revolution.
        direction : UnitVector3D
            Direction of extrusion.
        height : Distance | Quantity | Real,
            Height of the helix.
        pitch : Distance | Quantity | Real,
            Pitch of the helix.
        taper_angle : Angle | Quantity | Real,
            Taper angle of the helix.
        right_handed : bool,
            Right-handed helix if ``True``, left-handed if ``False``.
        both_sides : bool,
            Create on both sides if ``True``, one side if ``False``.
        extrude_type : ExtrudeType, default: ExtrudeType.ADD
            Type of extrusion to be performed.

        Returns
        -------
        list[Body]
            Bodies created by the extrusion if any.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        selection: list[Face] = selection if isinstance(selection, list) else [selection]
        check_type_all_elements_in_iterable(selection, Face)

        height = height if isinstance(height, Distance) else Distance(height)
        pitch = pitch if isinstance(pitch, Distance) else Distance(pitch)
        taper_angle = taper_angle if isinstance(taper_angle, Angle) else Angle(taper_angle)

        for object in selection:
            object.body._reset_tessellation_cache()

        result = self._grpc_client._services.faces.revolve_faces_by_helix(
            selection_ids=[object.id for object in selection],
            axis=axis,
            direction=direction,
            height=height,
            pitch=pitch,
            taper_angle=taper_angle,
            right_handed=right_handed,
            both_sides=both_sides,
            extrude_type=extrude_type,
        )

        design = get_design_from_face(selection[0])

        if result.get("success"):
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                design._update_from_tracker(result.get("tracked_response"))
            else:
                design._update_design_inplace()
            return get_bodies_from_ids(design, result.get("created_bodies"))
        else:
            self._grpc_client.log.info("Failed to revolve faces.")
            return []

    @min_backend_version(25, 2, 0)
    def replace_face(
        self,
        target_selection: Union["Face", list["Face"]],
        replacement_selection: Union["Face", list["Face"]],
    ) -> bool:
        """Replace a face with another face.

        Parameters
        ----------
        target_selection : Union[Face, list[Face]]
            The face or faces to replace.
        replacement_selection : Union[Face, list[Face]]
            The face or faces to replace with.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        target_selection: list["Face"] = (
            target_selection if isinstance(target_selection, list) else [target_selection]
        )
        replacement_selection: list["Face"] = (
            replacement_selection
            if isinstance(replacement_selection, list)
            else [replacement_selection]
        )

        result = self._grpc_client._services.faces.replace_faces(
            target_ids=[selection.id for selection in target_selection],
            replacement_ids=[selection.id for selection in replacement_selection],
        )

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def split_body(
        self,
        bodies: list["Body"],
        plane: Plane,
        slicers: Union["Edge", list["Edge"], "Face", list["Face"]],
        faces: list["Face"],
        extendfaces: bool,
    ) -> bool:
        """Split bodies with a plane, slicers, or faces.

        Parameters
        ----------
        bodies : list[Body]
            Bodies to split.
        plane : Plane
            Plane to split with
        slicers : Edge | list[Edge] | Face | list[Face]
            Slicers to split with.
        faces : list[Face]
            Faces to split with.
        extendFaces : bool
            Extend faces if split with faces.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.body import Body
        from ansys.geometry.core.designer.edge import Edge
        from ansys.geometry.core.designer.face import Face

        check_type_all_elements_in_iterable(bodies, Body)

        for body in bodies:
            body._reset_tessellation_cache()

        if plane is not None:
            check_type(plane, Plane)

        slicer_items = []
        if slicers is not None:
            slicers: list["Face", "Edge"] = slicers if isinstance(slicers, list) else [slicers]
            check_type_all_elements_in_iterable(slicers, (Edge, Face))
            slicer_items = [slicer.id for slicer in slicers]

        face_items = []
        if faces is not None:
            faces: list["Face"] = faces if isinstance(faces, list) else [faces]
            check_type_all_elements_in_iterable(faces, Face)
            face_items = [face.id for face in faces]

        result = self._grpc_client._services.bodies.split_body(
            body_ids=[body.id for body in bodies],
            plane=plane,
            slicer_ids=slicer_items,
            face_ids=face_items,
            extend_surfaces=extendfaces,
        )

        if result.get("success"):
            design = get_design_from_body(bodies[0])
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                design._update_from_tracker(result.get("tracked_response"))
            else:
                design._update_design_inplace()

        return result.get("success")

    @min_backend_version(25, 2, 0)
    def get_round_info(self, face: "Face") -> tuple[bool, Real]:
        """Get info on the rounding of a face.

        Parameters
        ----------
        Face
            The design face to get round info on.

        Returns
        -------
        tuple[bool, Real]
            ``True`` if round is aligned with face's U-parameter direction, ``False`` otherwise.
            Radius of the round.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        result = self._grpc_client._services.faces.get_round_info(face_id=face.id)

        return (result.get("along_u"), result.get("radius"))

    @check_input_types
    @min_backend_version(25, 2, 0)
    def move_translate(
        self,
        selection: NamedSelection,
        direction: UnitVector3D,
        distance: Distance | Quantity | Real,
    ) -> bool:
        """Move a selection by a distance in a direction.

        Parameters
        ----------
        selection : NamedSelection
            Named selection to move.
        direction : UnitVector3D
            Direction to move in.
        distance : Distance | Quantity | Real
            Distance to move. Default units are meters.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        distance = distance if isinstance(distance, Distance) else Distance(distance)

        result = self._grpc_client.services.model_tools.move_translate(
            selection_id=selection.id,
            direction=direction,
            distance=distance,
        )

        return result.get("success")

    @check_input_types
    @min_backend_version(25, 2, 0)
    def move_rotate(
        self,
        selection: NamedSelection,
        axis: Line,
        angle: Angle | Quantity | Real,
    ) -> dict[str, Union[bool, Real]]:
        """Rotate a selection by an angle about a given axis.

        Parameters
        ----------
        selection : NamedSelection
            Named selection to move.
        axis : Line
            Direction to move in.
        Angle : Angle | Quantity | Real
            Angle to rotate by. Default units are radians.

        Returns
        -------
        dict[str, Union[bool, Real]]
            Dictionary containing the useful output from the command result.
            Keys are success, modified_bodies, modified_faces, modified_edges.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        angle = angle if isinstance(angle, Angle) else Angle(angle)

        result = self._grpc_client.services.model_tools.move_rotate(
            selection_id=selection.id,
            axis=axis,
            angle=angle,
        )

        return result

    @check_input_types
    @min_backend_version(25, 2, 0)
    def offset_faces_set_radius(
        self,
        faces: Union["Face", list["Face"]],
        radius: Distance | Quantity | Real,
        copy: bool = False,
        offset_mode: OffsetMode = OffsetMode.IGNORE_RELATIONSHIPS,
        extrude_type: ExtrudeType = ExtrudeType.FORCE_INDEPENDENT,
    ) -> bool:
        """Offset faces with a radius.

        Parameters
        ----------
        faces : Face | list[Face]
            Faces to offset.
        radius : Distance | Quantity | Real
            Radius of the offset.
        copy : bool, default: False
            Copy the face and move it instead of offsetting the original face if ``True``.
        offset_mode : OffsetMode, default: OffsetMode.MOVE_FACES_TOGETHER
            Mode of how to handle offset relationships.
        extrude_type : ExtrudeType, default: ExtrudeType.FORCE_INDEPENDENT
            Type of extrusion to be performed.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.face import Face

        faces: list[Face] = faces if isinstance(faces, list) else [faces]
        check_type_all_elements_in_iterable(faces, Face)

        for face in faces:
            face.body._reset_tessellation_cache()

        radius = radius if isinstance(radius, Distance) else Distance(radius)

        result = self._grpc_client._services.faces.offset_faces_set_radius(
            face_ids=[face.id for face in faces],
            radius=radius,
            copy=copy,
            offset_mode=offset_mode,
            extrude_type=extrude_type,
        )

        return result.get("success")

    @min_backend_version(26, 1, 0)
    def create_align_condition(
        self,
        parent_component: "Component",
        geometry_a: Union["Body", "Face", "Edge"],
        geometry_b: Union["Body", "Face", "Edge"],
    ) -> AlignCondition:
        """Create an align condition between two geometry objects.

        This will move the objects to be aligned with each other.

        Parameters
        ----------
        parent_component : Component
            The common ancestor component of the two geometry objects.
        geometry_a : Body | Face | Edge
            The first geometry object to align to the second.
        geometry_b : Body | Face | Edge
            The geometry object to be aligned to.

        Returns
        -------
        AlignCondition
            The persistent align condition that was created.

        Warnings
        --------
        This method is only available starting on Ansys release 26R1.
        """
        from ansys.geometry.core.designer.body import Body
        from ansys.geometry.core.designer.component import Component
        from ansys.geometry.core.designer.edge import Edge
        from ansys.geometry.core.designer.face import Face

        check_type(parent_component, Component)
        check_type(geometry_a, (Body, Face, Edge))
        check_type(geometry_b, (Body, Face, Edge))

        result = self._grpc_client._services.assembly_condition.create_align_condition(
            parent_id=parent_component.id,
            geometric_a_id=geometry_a.id,
            geometric_b_id=geometry_b.id,
        )

        design = get_design_from_component(parent_component)
        design._update_design_inplace()

        return AlignCondition(
            result.get("id"),
            result.get("is_deleted"),
            result.get("is_enabled"),
            result.get("is_satisfied"),
            result.get("offset"),
            result.get("is_reversed"),
            result.get("is_valid"),
        )

    @min_backend_version(26, 1, 0)
    def create_tangent_condition(
        self,
        parent_component: "Component",
        geometry_a: Union["Body", "Face", "Edge"],
        geometry_b: Union["Body", "Face", "Edge"],
    ) -> TangentCondition:
        """Create a tangent condition between two geometry objects.

        This aligns the objects so that they are tangent.

        Parameters
        ----------
        parent_component : Component
            The common ancestor component of the two geometry objects.
        geometry_a : Body | Face | Edge
            The first geometry object to tangent the second.
        geometry_b : Body | Face | Edge
            The geometry object to be tangent with.

        Returns
        -------
        TangentCondition
            The persistent tangent condition that was created.

        Warnings
        --------
        This method is only available starting on Ansys release 26R1.
        """
        from ansys.geometry.core.designer.body import Body
        from ansys.geometry.core.designer.component import Component
        from ansys.geometry.core.designer.edge import Edge
        from ansys.geometry.core.designer.face import Face

        check_type(parent_component, Component)
        check_type(geometry_a, (Body, Face, Edge))
        check_type(geometry_b, (Body, Face, Edge))

        result = self._grpc_client._services.assembly_condition.create_tangent_condition(
            parent_id=parent_component.id,
            geometric_a_id=geometry_a.id,
            geometric_b_id=geometry_b.id,
        )

        design = get_design_from_component(parent_component)
        design._update_design_inplace()

        return TangentCondition(
            result.get("id"),
            result.get("is_deleted"),
            result.get("is_enabled"),
            result.get("is_satisfied"),
            result.get("offset"),
            result.get("is_reversed"),
            result.get("is_valid"),
        )

    @min_backend_version(26, 1, 0)
    def create_orient_condition(
        self,
        parent_component: "Component",
        geometry_a: Union["Body", "Face", "Edge"],
        geometry_b: Union["Body", "Face", "Edge"],
    ) -> OrientCondition:
        """Create an orient condition between two geometry objects.

        This rotates the objects so that they are oriented in the same direction.

        Parameters
        ----------
        parent_component : Component
            The common ancestor component of the two geometry objects.
        geometry_a : Body | Face | Edge
            The first geometry object to orient with the second.
        geometry_b : Body | Face | Edge
            The geometry object to be oriented with.

        Returns
        -------
        OrientCondition
            The persistent orient condition that was created.

        Warnings
        --------
        This method is only available starting on Ansys release 26R1.
        """
        from ansys.geometry.core.designer.body import Body
        from ansys.geometry.core.designer.component import Component
        from ansys.geometry.core.designer.edge import Edge
        from ansys.geometry.core.designer.face import Face

        check_type(parent_component, Component)
        check_type(geometry_a, (Body, Face, Edge))
        check_type(geometry_b, (Body, Face, Edge))

        result = self._grpc_client.services.assembly_condition.create_orient_condition(
            parent_id=parent_component.id,
            geometric_a_id=geometry_a.id,
            geometric_b_id=geometry_b.id,
        )

        design = get_design_from_component(parent_component)
        design._update_design_inplace()

        return OrientCondition(
            result.get("id"),
            result.get("is_deleted"),
            result.get("is_enabled"),
            result.get("is_satisfied"),
            result.get("offset"),
            result.get("is_reversed"),
            result.get("is_valid"),
        )

    @min_backend_version(26, 1, 0)
    def move_imprint_edges(
        self, edges: list["Edge"], direction: UnitVector3D, distance: Distance | Quantity | Real
    ) -> bool:
        """Move the imprint edges in the specified direction by the specified distance.

        Parameters
        ----------
        edges : list[Edge]
            The edges to move.
        direction : UnitVector3D
            The direction to move the edges.
        distance : Distance | Quantity | Real
            The distance to move the edges.

        Returns
        -------
        bool
            Returns True if the edges were moved successfully, False otherwise.
        """
        distance = distance if isinstance(distance, Distance) else Distance(distance)

        response = self._grpc_client._services.edges.move_imprint_edges(
            edge_ids=[edge.id for edge in edges],
            direction=direction,
            distance=distance,
        )

        return response.get("success")

    @min_backend_version(26, 1, 0)
    def offset_edges(self, edges: list["Edge"], offset: Distance | Quantity | Real) -> bool:
        """Offset the specified edges with the specified distance.

        Parameters
        ----------
        edges : list[Edge]
            The edges to offset.
        offset : Distance | Quantity | Real
            The distance to offset the edges.

        Returns
        -------
        bool
            Returns True if the edges were offset successfully, False otherwise.
        """
        offset = offset if isinstance(offset, Distance) else Distance(offset)

        response = self._grpc_client._services.edges.offset_edges(
            edge_ids=[edge.id for edge in edges],
            offset=offset,
        )

        return response.get("success")

    @min_backend_version(26, 1, 0)
    def draft_faces(
        self,
        faces: list["Face"],
        reference_faces: list["Face"],
        draft_side: DraftSide,
        angle: Angle | Quantity | Real,
        extrude_type: ExtrudeType,
    ) -> list["Face"]:
        """Draft the specified faces in the specified direction by the specified angle.

        Parameters
        ----------
        faces : list[Face]
            The faces to draft.
        reference_faces : list[Face]
            The reference faces to use for the draft.
        draft_side : DraftSide
            The side to draft.
        angle : Angle | Quantity | Real
            The angle to draft the faces.
        extrude_type : ExtrudeType
            The type of extrusion to use.

        Returns
        -------
        list[Face]
            The faces created by the draft operation.
        """
        angle = angle if isinstance(angle, Angle) else Angle(angle)

        response = self._grpc_client._services.faces.draft_faces(
            face_ids=[face.id for face in faces],
            reference_face_ids=[face.id for face in reference_faces],
            draft_side=draft_side,
            angle=angle,
            extrude_type=extrude_type,
        )

        # Return the drafted faces
        design = get_design_from_face(faces[0])
        return get_faces_from_ids(design, [face.id for face in response.get("created_faces")])

    @min_backend_version(26, 1, 0)
    def thicken_faces(
        self,
        faces: list["Face"],
        direction: UnitVector3D,
        thickness: Distance | Quantity | Real,
        extrude_type: ExtrudeType,
        pull_symmetric: bool,
        select_direction: bool,
    ) -> bool:
        """Thicken the specified faces by the specified thickness in the specified direction.

        Parameters
        ----------
        faces : list[Face]
            The faces to thicken.
        direction : UnitVector3D
            The direction to thicken the faces.
        thickness : Distance | Quantity | Real
            The thickness to apply to the faces.
        extrude_type : ExtrudeType
            The type of extrusion to use.
        pull_symmetric : bool
            Whether to pull the faces symmetrically.
        select_direction : bool
            Whether to select the direction.

        Returns
        -------
        bool
            Returns True if the faces were thickened successfully, False otherwise.
        """
        thickness = thickness if isinstance(thickness, Distance) else Distance(thickness)

        result = self._grpc_client._services.faces.thicken_faces(
            face_ids=[face.id for face in faces],
            direction=direction,
            thickness=thickness,
            extrude_type=extrude_type,
            pull_symmetric=pull_symmetric,
            select_direction=select_direction,
        )

        # Update design
        design = get_design_from_face(faces[0])
        if result.get("success"):
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                design._update_from_tracker(result.get("tracked_response"))
            else:
                design._update_design_inplace()

        # Return success flag
        return result.get("success")

    @min_backend_version(26, 1, 0)
    def offset_faces(
        self,
        faces: list["Face"],
        distance: Distance | Quantity | Real,
        direction: UnitVector3D,
        extrude_type: ExtrudeType,
    ) -> None:
        """Offset the specified faces by the specified distance in the specified direction.

        Parameters
        ----------
        faces : list[Face]
            The faces to offset.
        distance : Distance | Quantity | Real
            The distance to offset the faces.
        direction : UnitVector3D
            The direction to offset the faces.
        extrude_type : ExtrudeType
            The type of extrusion to use.

        Warnings
        --------
        This method is only available starting on Ansys release 26R1.
        """
        distance = distance if isinstance(distance, Distance) else Distance(distance)

        _ = self._grpc_client._services.faces.offset_faces(
            face_ids=[face.id for face in faces],
            distance=distance,
            direction=direction,
            extrude_type=extrude_type,
        )

    @min_backend_version(25, 2, 0)
    def revolve_edges(
        self,
        edges: Union["Edge", list["Edge"]],
        axis: Line,
        angle: Angle | Quantity | Real,
        symmetric: bool,
    ) -> None:
        """Revolve edges around an axis.

        Parameters
        ----------
        edges : Edge | list[Edge]
            Edge(s) to revolve.
        axis : Line
            Axis of revolution.
        angle : Angle | Quantity | Real
            Angular distance to revolve.
        symmetric : bool
            Revolve symmetrically if ``True``, one side if ``False``.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        from ansys.geometry.core.designer.edge import Edge

        edges: list[Edge] = edges if isinstance(edges, list) else [edges]
        check_type_all_elements_in_iterable(edges, Edge)

        angle = angle if isinstance(angle, Angle) else Angle(angle)

        result = self._grpc_client._services.curves.revolve_edges(
            curves=[edge.shape for edge in edges],
            axis=axis,
            angle=angle,
            symmetric=symmetric,
        )

        design = get_design_from_edge(edges[0])
        if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
            design._update_from_tracker(result.get("tracked_response"))
        else:
            design._update_design_inplace()

    @min_backend_version(26, 1, 0)
    def intersect_curve_and_surface(
        self,
        curve: Curve,
        surface: Surface,
    ) -> list[Point3D]:
        """Find the intersection points of a curve and a surface.

        Parameters
        ----------
        curve : Curve
            Curve to intersect.
        surface : Surface
            Surface to intersect.

        Returns
        -------
        list[Point3D]
            Points of intersection.

        Warnings
        --------
        This method is only available starting on Ansys release 26R1. Nurbs curves
        and surfaces are not supported until Ansys release 27R1.
        """
        response = self._grpc_client._services.curves.intersect_curve_and_surface(
            curve=curve, surface=surface
        )

        return [] if response.get("intersect") is False else response.get("points")

    @min_backend_version(27, 1, 0)
    def detach_faces(
        self,
        selection: Union["Body", list["Body"], "Face", list["Face"]],
    ) -> list["Body"]:
        """Detach faces on all the bodies/faces from a list.

        This method will result in a list of new surface bodies:

        - If the input is a body, all faces on the body will be detached and the original body
          will be modified to have only the non-detached faces. The detached faces will be
          returned as new bodies.
        - If the input is a list of faces:
            - All the connected faces from the same body will be detached together to form a
              body.
            - If some of the faces in the list are not connected or are not from the same body,
              they will be detached separately to form separate bodies.
            - The original body will be modified to have only the non-detached faces.
              The detached faces will be returned as new bodies.

        Parameters
        ----------
        selection : Body | list[Body] | Face | list[Face]
            Bodies or faces from which we want to detach faces.

        Returns
        -------
        list[Body]
            Bodies created by the detach if any.

        Warnings
        --------
        This method is only available starting on Ansys release 27R1.
        """
        from ansys.geometry.core.designer.face import Face

        selection_ids: list[str] = (
            [entity.id for entity in selection] if isinstance(selection, list) else [selection.id]
        )
        check_type_all_elements_in_iterable(selection_ids, str)

        response = self._grpc_client.services.model_tools.detach_faces(selections=[selection_ids])

        first_item = selection[0] if isinstance(selection, list) else selection
        parent_design = (
            get_design_from_face(first_item)
            if isinstance(first_item, Face)
            else get_design_from_body(first_item)
        )

        if response.get("success"):
            if pyansys_geo.USE_TRACKER_TO_UPDATE_DESIGN:
                parent_design._update_from_tracker(response.get("tracked_response"))
            else:
                parent_design._update_design_inplace()

            result_bodies = response.get("created_bodies")
            return get_bodies_from_ids(parent_design, result_bodies)
        else:
            self._grpc_client.log.info("Failed to detach faces.")
            return []
